import json
import logging
import time
from typing import cast, Dict, List, Literal, Union
from uuid import UUID

from pydantic import Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EntityType, File, Response, ResponseData
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.entities.parallel_experiment.row import Row
from signals_notebook.jinja_env import env

log = logging.getLogger(__name__)


class SubExpSummaryResponse(Response[Row]):
    pass


class SubExperimentSummary(ContentfulEntity):
    type: Literal[EntityType.SUB_EXPERIMENT_SUMMARY] = Field(allow_mutation=False)
    _rows: List[Row] = PrivateAttr(default=[])
    _rows_by_id: Dict[Union[str, UUID], Row] = PrivateAttr(default={})
    _template_name = 'subexp_summary.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.SUB_EXPERIMENT_SUMMARY

    @classmethod
    def _get_subexpsum_endpoint(cls) -> str:
        return 'subexpSummary'

    def __getitem__(self, index: Union[int, str, UUID]) -> Row:
        if not self._rows:
            self._reload_cells()

        if isinstance(index, int):
            return self._rows[index]

        if isinstance(index, str):
            try:
                return self._rows_by_id[UUID(index)]
            except ValueError:
                return self._rows_by_id[index]

        if isinstance(index, UUID):
            return self._rows_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._rows:
            self._reload_cells()
        return self._rows.__iter__()

    def _reload_cells(self) -> None:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Reloading data in Table: %s...', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_subexpsum_endpoint(), self.eid, 'rows'),
        )

        result = SubExpSummaryResponse(**response.json())  # type: ignore
        cells = [cast(ResponseData, item).body for item in result.data]

        self._rows = []
        self._rows_by_id = {}

        for item in cells:
            subexp_row = cast(Row, item)
            assert subexp_row.id

            self._rows.append(subexp_row)
            self._rows_by_id[subexp_row.id] = subexp_row

    def _is_update_ready(self, update_id: str) -> bool:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Check job status for: %s| %s', self.__class__.__name__, self.eid)

        response = api.call(
            method='GET',
            path=(self._get_subexpsum_endpoint(), self.eid, 'bulkUpdate', update_id),
        )
        return response.status_code == 200 and response.json()['data']['attributes']['jobStatus'] == 'SUCCESS'

    def save(self, force: bool = True, timeout: int = 30, period: int = 5) -> None:
        """Save all changes in the table

        Args:
            force: Force to update properties without digest check.
            timeout: max available time(seconds) to update table
            period: each n seconds(default value=5) api call

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()

        request_body = []
        for item in self._rows:
            if item.is_changed:
                request_body.extend(item.representation_for_update)

        if not request_body:
            return

        update_response = api.call(
            method='PATCH',
            path=(self._get_subexpsum_endpoint(), self.eid, 'bulkUpdate'),
            params={
                'force': json.dumps(force),
            },
            json={
                'data': request_body,
            },
        )
        bulk_update_id = update_response.json()['data']['attributes']['bulkUpdateId']

        response = None
        initial_time = time.time()
        while time.time() - initial_time < timeout:
            if self._is_update_ready(str(bulk_update_id)):
                response = True
                break
            else:
                time.sleep(period)

        if not response:
            log.debug('Time is over to update fields')

        self._reload_cells()

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered HTML in string format
        """
        if not self._rows:
            self._reload_cells()

        table_head = None
        if self._rows:
            table_head = self._rows[0]
        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(name=self.name, table_head=table_head, rows=self._rows)

    def get_content(self) -> File:
        """Get SubExperiment Summary content

        Returns:
            File
        """
        return super()._get_content(format='csv')



from typing import Generic, List, Optional, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr
from pydantic.generics import GenericModel

from signals_notebook.common_types import EID, ObjectType

CellValueType = TypeVar('CellValueType')


class TaskCellContent(GenericModel, Generic[CellValueType]):
    value: Optional[CellValueType]
    name: Optional[str]
    eid: Optional[EID]
    values: Optional[List[CellValueType]]
    _changed: bool = PrivateAttr(default=False)

    def set_value(self, new_value: CellValueType) -> None:
        """Set new value

        Args:
            new_value: new value of Property value field

        Returns:

        """
        self.value = new_value
        self._changed = True

    def set_values(self, new_values: List[CellValueType]) -> None:
        """Set new values

        Args:
            new_values: new list of values of Property values field

        Returns:

        """
        self.values = new_values
        self._changed = True

    def set_name(self, new_name: str) -> None:
        """Set new name

        Args:
            new_name: new name of Property name field

        Returns:

        """
        self.name = new_name
        self._changed = True

    @property
    def is_changed(self) -> bool:
        """Checking if content of Cell has been modified

        Returns:
            bool
        """
        return self._changed


class Content(BaseModel):
    content: Optional[TaskCellContent]


class TaskCellBody(BaseModel):
    id: Optional[Union[UUID, str]]
    type: str = ObjectType.PROPERTY
    attributes: Content


class TaskCell(BaseModel):
    id: Optional[Union[UUID, str]]
    name: Optional[str]
    content: TaskCellContent = Field(default=TaskCellContent())

    def set_content_value(self, new_value: CellValueType) -> None:
        """Set new content value

        Args:
            new_value: new value of Property value field

        Returns:

        """
        self.content.set_value(new_value)

    def set_content_values(self, new_values: List[CellValueType]) -> None:
        """Set new content values

        Args:
            new_values: new list of values of Property values field

        Returns:

        """
        self.content.set_values(new_values)

    def set_content_name(self, new_name: str) -> None:
        """Set new content name

        Args:
            new_name: new name of content name field

        Returns:

        """
        self.content.set_name(new_name)

    @property
    def content_value(self) -> Optional[CellValueType]:
        """Get content value

        Returns:
            Optional[CellValueType]
        """
        return self.content.value

    @property
    def content_values(self) -> Optional[List[CellValueType]]:
        """Get content values

        Returns:
            Optional[List[CellValueType]]
        """
        return self.content.values

    @property
    def content_name(self) -> Optional[str]:
        """Get content name

        Returns:
            Optional[str]
        """
        return self.content.name

    @property
    def is_changed(self) -> bool:
        """Checking if Property has been modified

        Returns:
            bool
        """
        return False if self.content is None else self.content.is_changed

    @property
    def representation_for_update(self) -> TaskCellBody:
        """Get representation of body for update

        Returns:
            TaskCellBody
        """
        return TaskCellBody(id=str(self.id), attributes=Content(content=self.content))



import cgi
import json
import logging
import mimetypes
from datetime import datetime
from typing import Any, cast, Generator, Optional, TYPE_CHECKING, Union

from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import File, Response, ResponseData
from signals_notebook.users.role import Role

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from signals_notebook.users.group import Group


class Licence(BaseModel):
    id: Union[int, str]
    name: str
    expires_at: datetime = Field(alias='expiresAt')
    valid: bool
    has_service_expired: bool = Field(alias='hasServiceExpired')
    has_user_found: bool = Field(alias='hasUserFound')
    has_user_activated: bool = Field(alias='hasUserActivated')

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True


class BaseUser(BaseModel):
    id: str = Field(alias='userId', allow_mutation=False)
    first_name: str = Field(alias='firstName')
    last_name: str = Field(alias='lastName')
    email: str = Field(allow_mutation=False)
    created_at: datetime = Field(alias='createdAt', allow_mutation=False)

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True


class Profile(BaseUser):
    tenant: str
    roles: list[Role]
    licenses: list[Licence]


class ProfileResponse(Response[Profile]):
    pass


class User(BaseUser):
    is_enabled: Optional[bool] = Field(alias='isEnabled', allow_mutation=False)
    username: str = Field(alias='userName', allow_mutation=False)
    alias: Optional[str] = Field(alias='alias')
    country: str = Field(alias='country')
    organization: str = Field(alias='organization')
    last_login_at: Optional[datetime] = Field(alias='lastLoginAt', allow_mutation=False)
    _picture: Optional[File] = PrivateAttr(default=None)
    _groups: list['Group'] = PrivateAttr(default=[])
    _roles: list[Role] = PrivateAttr(default=[])
    _relationships: dict[str, Any] = PrivateAttr(default={})

    def set_relationships(self, value: dict) -> None:
        self._relationships = value

    @property
    def roles(self) -> list[Role]:
        if self._roles:
            return self._roles
        self._roles = [Role.get(role['id']) for role in self._relationships['roles']['data']]
        return self._roles

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'users'

    @staticmethod
    def get(user_id: str) -> 'User':
        """Get user by id from the scope

        Returns:
            User
        """
        log.debug('Get user: %s', user_id)

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=('users', user_id),
        )
        result = UserResponse(**response.json())

        user = cast(ResponseData, result.data).body
        user.set_relationships(result.data.relationships)  # type: ignore

        log.debug('Role: %s was got successfully.', user_id)

        return user

    @staticmethod
    def get_list(q: str = '', enabled: bool = True, offset: int = 0, limit: int = 20) -> Generator['User', None, None]:
        """Get all users from the scope

        Parameter 'q' is a String and it is used to filter users.

        Example usage: Input 'foo' as value of 'q', will return
        Users whose first name starts with 'foo'.
        Users whose last name starts with 'foo'.
        Users whose email address starts with 'foo'.
        Users who has role name (except "Standard User") contains 'foo'.
        All of above are case insensitive.

        Args:
            q: query string for list users
            enabled: filter activated and deactivated users
            offset: Number of items to skip before returning the results.
            limit: Maximum number of items to return.

        Returns:
            User
        """
        api = SignalsNotebookApi.get_default_api()

        log.debug('Get List of Users')

        response = api.call(
            method='GET',
            path=('users',),
            params={
                'q': q,
                'enabled': json.dumps(enabled),
                'offset': offset,
                'limit': limit,
            },
        )
        result = UserResponse(**response.json())

        for item in result.data:
            user = cast(ResponseData, item).body
            user.set_relationships(item.relationships)  # type: ignore

            yield user

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = UserResponse(**response.json())

            for item in result.data:
                user = cast(ResponseData, item).body
                user.set_relationships(item.relationships)  # type: ignore

                yield user

        log.debug('List of Users was got successfully.')

    @staticmethod
    def get_current_user() -> Profile:
        """Get current user for api session

        Returns:
            User
        """
        api = SignalsNotebookApi.get_default_api()

        log.debug('Get current user')

        response = api.call(
            method='GET',
            path=('profiles', 'me'),
        )
        result = ProfileResponse(**response.json())

        log.debug('Current user was got successfully.')

        return cast(ResponseData, result.data).body

    @classmethod
    def create(
        cls,
        alias: str,
        country: str,
        email: str,
        first_name: str,
        last_name: str,
        organization: str,
        roles: Optional[list[Role]],
    ) -> 'User':
        """Create new User

        Returns:
            User
        """
        api = SignalsNotebookApi.get_default_api()

        log.debug('Create User: %s...', cls.__name__)

        response = api.call(
            method='POST',
            path=(cls._get_endpoint(),),
            json={
                'data': {
                    'attributes': {
                        'alias': alias,
                        'country': country,
                        'emailAddress': email,
                        'firstName': first_name,
                        'lastName': last_name,
                        'organization': organization,
                        'roles': [role.dict(include={'id', 'name'}) for role in roles] if roles else None,
                    },
                }
            },
        )

        result = UserResponse(**response.json())

        user = cast(ResponseData, result.data).body
        user.set_relationships(result.data.relationships)  # type: ignore

        log.debug('User: %s was created.', cls.__name__)

        return user

    def refresh(self) -> None:
        """Refresh user with new changes values

        Returns:

        """
        refreshed_entity = self.get(self.id)
        for field in self.__fields__.values():
            if field.field_info.allow_mutation:
                new_value = getattr(refreshed_entity, field.name)
                setattr(self, field.name, new_value)

    def save(self) -> None:
        """Update attributes of a specified user.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        api.call(
            method='PATCH',
            path=(
                self._get_endpoint(),
                self.id,
            ),
            json={
                'data': {
                    'attributes': self.dict(
                        by_alias=True,
                        include={
                            'alias',
                            'country',
                            'first_name',
                            'last_name',
                            'organization',
                        },
                    )
                },
            },
        )
        log.debug('User: %s was saved successfully', self.id)

    def delete(self) -> None:
        """Make specified user disabled.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Disable User: %s...', self.id)

        api.call(
            method='DELETE',
            path=(self._get_endpoint(), self.id),
        )
        log.debug('User: %s was disabled successfully', self.id)

    @property
    def picture(self) -> Optional[File]:
        if self._picture:
            return self._picture

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.id, 'picture'),
        )
        if response.content == b'':
            return self._picture

        content_disposition = response.headers.get('content-disposition', '')
        content_type = response.headers.get('content-type', '')
        extension = mimetypes.guess_extension(content_type)

        _, params = cgi.parse_header(content_disposition)
        file_name = f'{self.first_name}_{self.last_name}{extension}'
        self._picture = File(name=file_name, content=response.content, content_type=content_type)
        return self._picture

    @property
    def groups(self) -> list['Group']:
        from signals_notebook.users.group import GroupResponse

        if self._groups:
            return self._groups

        api = SignalsNotebookApi.get_default_api()

        log.debug('Getting System Groups')

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.id, 'systemGroups'),
        )
        result = GroupResponse(**response.json())

        log.debug('List of groups was got successfully.')

        return [cast(ResponseData, item).body for item in result.data]


class UserResponse(Response[User]):
    pass



import json
import logging
from datetime import datetime
from enum import Enum
from typing import cast, Generator, List, Optional, Union

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EID, EntityType, Response, ResponseData
from signals_notebook.entities import Entity
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class EntityStore:
    class IncludeOptions(str, Enum):
        MINE = 'mine'
        OTHER = 'other'
        SHARED = 'shared'
        TRASHED = 'trashed'
        UNTRASHED = 'untrashed'
        TRASHED_ANCESTOR = 'trashedAncestor'
        STARRED = 'starred'
        UNSTARRED = 'unstarred'
        TEMPLATE = 'template'
        NONTEMPLATE = 'nontemplate'
        SYSTEM_TEMPLATE = 'systemTemplate'
        NON_SYSTEM_TEMPLATE = 'nonSystemTemplate'

    @staticmethod
    def _get_endpoint() -> str:
        return 'entities'

    @classmethod
    def get(cls, eid: EID) -> Entity:
        """Get Entity by ID

        Args:
            eid: Entity ID

        Returns:
            Entity
        """

        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Entity: %s from EntityStore...', eid)

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), eid),
        )

        entity_classes = (*Entity.get_subclasses(), Entity)
        result = Response[Union[entity_classes]](**response.json())  # type: ignore
        log.debug('Entity: %s was got successfully from EntityStore.', eid)

        return cast(ResponseData, result.data).body

    @classmethod
    def get_list(
        cls,
        include_types: Optional[List[EntityType]] = None,
        exclude_types: Optional[List[EntityType]] = None,
        include_options: Optional[List[IncludeOptions]] = None,
        modified_after: Optional[datetime] = None,
        modified_before: Optional[datetime] = None,
    ) -> Generator[Entity, None, None]:
        """Get all entities

        Args:
            include_types: Included entity types, separated by comma ','.
                For example, 'experiment, journal, ado'.
                The default include types are experiment, request and journal.
            exclude_types: Excluded entity types, separated by comma ','. For example, 'experiment, journal'.
            include_options: Flags of entities, separated by comma ','.
            modified_after: Return the entities which are modified after start time.
            modified_before: Return the entities which are modified before end time.

        Returns:
            Entity
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get List of Entities from EntityStore...')

        params = {}
        if include_types:
            params['includeTypes'] = ','.join(include_types)
        if exclude_types:
            params['excludeTypes'] = ','.join(exclude_types)
        if include_options:
            params['includeOptions'] = ','.join(include_options)
        if modified_after:
            params['start'] = modified_after.isoformat()
        if modified_before:
            params['end'] = modified_before.isoformat()

        entity_classes = (*Entity.get_subclasses(), Entity)

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(),),
            params=params or None,
        )

        result = Response[Union[entity_classes]](**response.json())  # type: ignore
        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = Response[Union[entity_classes]](**response.json())  # type: ignore
            yield from [cast(ResponseData, item).body for item in result.data]

        log.debug('List of Entities were got successfully from EntityStore.')

    @classmethod
    def refresh(cls, entity: Entity) -> None:
        """Refresh Entity with new values

        Args:
            entity: Entity

        Returns:

        """
        refreshed_entity = cls.get(entity.eid)
        for field in entity.__fields__.values():
            if field.field_info.allow_mutation:
                new_value = getattr(refreshed_entity, field.name)
                setattr(entity, field.name, new_value)

    @classmethod
    def delete(cls, eid: EID, digest: Optional[str] = None, force: bool = True) -> None:
        """Delete Entity by ID

        Args:
            eid: Entity ID
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
                If the parameter 'force' is true, this parameter is optional.
                If the parameter 'force' is false, this parameter is required.
            force: Force to delete without doing digest check.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Deleting Entity: %s from EntityStore...', eid)

        api.call(
            method='DELETE',
            path=(cls._get_endpoint(), eid),
            params={
                'digest': digest,
                'force': json.dumps(force),
            },
        )
        log.debug('Entity: %s was deleted from EntityStore successfully', eid)

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump all templates from system

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """

        for item in Entity.get_subclasses():
            try:
                item.dump_templates(base_path, fs_handler)
            except Exception as e:
                log.error('Failed to dump templates for %s with error %s' % (str(item), str(e)))



from signals_notebook.materials.material_store import MaterialStore  # noqa
from signals_notebook.materials.library import Library  # noqa
from signals_notebook.materials.asset import Asset  # noqa
from signals_notebook.materials.batch import Batch  # noqa



import abc
import cgi
import json
import logging
from enum import Enum
from typing import Any, List, Optional

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EntityType, File
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class ContentfulEntity(Entity, abc.ABC):
    class ContentType(str, Enum):
        BYTES = 'application/octet-stream'

    def get_content(self) -> File:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def _get_entity_type(cls) -> EntityType:
        pass

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content_type: str = ContentType.BYTES,
        content: bytes = b'',
        force: bool = True,
    ) -> Entity:
        raise NotImplementedError

    def _get_content(self, format: Optional[str] = None) -> File:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get content for: %s| %s', self.__class__.__name__, self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'export'),
            params={
                'format': format,
            },
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'],
            content=response.content,
            content_type=response.headers.get('content-type'),
        )

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        file = self._get_content()
        data = {
            'name': self.name,
            'content': 'data:{};base64,{}'.format(file.content_type, file.base64.decode('ascii')),
        }
        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump ContentfulEntity entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        content = self.get_content()
        metadata = {
            'file_name': content.name,
            'content_type': content.content_type,
            **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
        }
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            base_alias=alias + [self.name, '__Metadata'] if alias else None,
        )
        file_name = content.name
        data = content.content
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, file_name),
            data,
            base_alias=alias + [self.name, file_name] if alias else None,
        )

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, parent: Container) -> None:
        """Load ContentfulEntity entity

        Args:
            path: content path
            fs_handler: FSHandler
            parent: Container where load ContentfulEntity entity

        Returns:

        """
        cls._load(path, fs_handler, parent)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        metadata_path = fs_handler.join_path(path, 'metadata.json')
        metadata = json.loads(fs_handler.read(metadata_path))
        content_path = fs_handler.join_path(path, metadata['file_name'])
        content_type = metadata.get('content_type')
        content = fs_handler.read(content_path)
        if content_type:
            cls.create(container=parent, name=metadata['name'], content=content, content_type=content_type, force=True)
        else:
            cls.create(container=parent, name=metadata['name'], content=content, force=True)



import cgi
import json
import logging
from typing import Any, cast, Optional, TYPE_CHECKING

from pydantic import PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import ChemicalDrawingFormat, File, MaterialType, MID
from signals_notebook.materials.base_entity import BaseMaterialEntity
from signals_notebook.materials.field import FieldContainer

if TYPE_CHECKING:
    from signals_notebook.materials.library import Library

log = logging.getLogger(__name__)


class Material(BaseMaterialEntity):

    _material_fields: FieldContainer = PrivateAttr(default={})
    _library: Optional['Library'] = PrivateAttr(default=None)

    def __init__(self, _library: Optional['Library'] = None, **data):
        super().__init__(**data)
        self._library = _library

    def __getitem__(self, key: str) -> Any:
        return self._material_fields[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._material_fields[key] = value

    @property
    def library(self) -> 'Library':
        """Fetch material library.

        Returns:
            Library
        """
        if not self._library:
            from signals_notebook.materials.material_store import MaterialStore

            library = MaterialStore.get(MID(f'{MaterialType.LIBRARY}:{self.asset_type_id}'))
            self._library = cast('Library', library)

        return self._library

    def get_chemical_drawing(self, format: Optional[ChemicalDrawingFormat] = None) -> File:
        """Export chemical drawing or image of a specified material.

        Args:
            format: Output type of chemical drawing.

        Returns:
            File
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Chemical Drawing as File for %s', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'drawing'),
            params={
                'format': format,
            },
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'], content=response.content, content_type=response.headers.get('content-type')
        )

    def get_image(self) -> File:
        """Export image of a specified material except Compounds/Reagents (SNB).

        Returns:
            File
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Image as File for %s', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'image'),
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'], content=response.content, content_type=response.headers.get('content-type')
        )

    def get_bio_sequence(self) -> File:
        """Export biological sequence file of a specified material with type is DNA or Protein."

        Returns:
            File
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Bio Sequence as File for %s', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'bioSequence'),
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'], content=response.content, content_type=response.headers.get('content-type')
        )

    def get_attachment(self, field_id: str) -> File:
        """Export an attachment for a specified field of the specific material.

        Args:
            field_id: Unique material field identifier.

        Returns:
            File
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Attachment as File for %s', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'attachments', field_id),
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'], content=response.content, content_type=response.headers.get('content-type')
        )

    def save(self, force: bool = True) -> None:
        """Update properties of a specified material.

        Args:
            force: Force to update properties without digest check.

        Returns:

        """
        request_body = []

        for field_name, field in self._material_fields.items():
            if field.is_changed:
                request_body.append(
                    {
                        'attributes': {
                            'name': field_name,
                            'value': field.value,
                        }
                    }
                )
        log.debug('Save %s: %s', self.__class__.__name__, self.eid)

        api = SignalsNotebookApi.get_default_api()

        api.call(
            method='PATCH',
            path=(self._get_endpoint(), self.eid, 'properties'),
            params={
                'digest': None if force else self.digest,
                'force': json.dumps(force),
            },
            json={
                'data': request_body,
            },
        )

    def delete(self, digest: Optional[str] = None, force: bool = True) -> None:
        """Delete Material by ID

        Args:
            eid: material ID
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
                If the parameter 'force' is true, this parameter is optional.
                If the parameter 'force' is false, this parameter is required.
            force: Force to delete without doing digest check.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Deleting Material: %s', self.eid)

        api.call(
            method='DELETE',
            path=('entities', self.eid),
            params={
                'digest': digest,
                'force': json.dumps(force),
            },
        )
        log.debug('Library: %s was deleted successfully', self.eid)



import json
import logging
from json import JSONDecodeError
from typing import cast, ClassVar, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import Field, PrivateAttr

from signals_notebook.common_types import EID, EntityType, File
from signals_notebook.entities import EntityStore, Task
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class TodoList(ContentfulEntity):
    type: Literal[EntityType.TODO_LIST] = Field(allow_mutation=False)
    _template_name: ClassVar = 'todo_list.html'
    _tasks: List[Task] = PrivateAttr(default=[])
    _tasks_by_id: Dict[EID, Task] = PrivateAttr(default={})

    def __getitem__(self, index: Union[int, str, UUID, EID]) -> Task:
        if not self._tasks:
            self._reload_tasks()

        if isinstance(index, int):
            return self._tasks[index]

        if isinstance(index, str):
            return self._tasks_by_id[EID(index)]

        if isinstance(index, EID):
            return self._tasks_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._tasks:
            self._reload_tasks()
        return self._tasks.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.TODO_LIST

    def _reload_tasks(self) -> None:
        log.debug('Reloading tasks in TodoList: %s...', self.eid)
        self._tasks = []
        self._tasks_by_id = {}
        file = self.get_content()
        content = file.content.decode('utf-8')
        try:
            dict_content = json.loads(content)
        except JSONDecodeError:
            return
        for item in dict_content['rows']:
            row = EntityStore.get(item['eid'])
            task = cast(Task, row)
            assert task.eid

            self._tasks.append(task)
            self._tasks_by_id[task.eid] = task
        log.debug('Tasks in TodoList: %s were reloaded', self.eid)

    def save(self, force: bool = True) -> None:
        """Save all changes in the TodoList

        Args:
            force: Force to update properties without digest check.

        Returns:

        """

        log.debug('Saving TodoList: %s...', self.eid)
        for item in self._tasks:
            item.save(force=force)
        self._reload_tasks()
        log.debug('TodoList: %s was saved successfully', self.eid)

    def get_content(self) -> File:
        """Get TodoList content

        Returns:
            File
        """
        return super()._get_content()

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """

        file = self.get_content()
        template = env.get_template(self._template_name)
        content = file.content.decode('utf-8')
        try:
            dict_content = json.loads(content)
        except json.JSONDecodeError:
            return template.render(name=self.name)

        table_head = dict_content['cols']
        rows = dict_content['rows']
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(name=self.name, table_head=table_head, rows=rows)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump TodoList entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler

        Returns:

        """
        super().dump(base_path=base_path, fs_handler=fs_handler, alias=alias + [self.name] if alias else None)
        for item in self:
            tasks_path = fs_handler.join_path(base_path, self.eid)
            item.dump(base_path=tasks_path, fs_handler=fs_handler, alias=alias + [self.name] if alias else None)



import json
import logging
from typing import cast, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import (
    EntityType,
    Response,
    ResponseData,
)
from signals_notebook.entities import Entity
from signals_notebook.entities.todo_list.cell import TaskCell
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class TaskCellsResponse(Response[TaskCell]):
    pass


class Task(Entity):
    type: Literal[EntityType.TASK] = Field(allow_mutation=False)
    _cells: List[TaskCell] = PrivateAttr(default=[])
    _cells_by_id: Dict[Union[str, UUID], TaskCell] = PrivateAttr(default={})

    def __getitem__(self, index: Union[int, str, UUID]) -> TaskCell:
        if not self._cells:
            self._reload_cells()

        if isinstance(index, int):
            return self._cells[index]

        if isinstance(index, str):
            try:
                return self._cells_by_id[UUID(index)]
            except ValueError:
                return self._cells_by_id[index]

        if isinstance(index, UUID):
            return self._cells_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._cells:
            self._reload_cells()
        return self._cells.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.TASK

    @classmethod
    def _get_tasks_endpoint(cls) -> str:
        return 'tasks'

    def _reload_cells(self) -> None:
        log.debug('Reloading cells in Task: %s...', self.eid)
        self._cells = []
        self._cells_by_id = {}

        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(self._get_tasks_endpoint(), self.eid, 'properties'),
            params={'value': 'normalized'},
        )

        result = TaskCellsResponse(**response.json())
        cells = [cast(ResponseData, item).body for item in result.data]

        for item in cells:
            task_cell = cast(TaskCell, item)
            assert task_cell.id

            self._cells.append(task_cell)
            self._cells_by_id[task_cell.id] = task_cell
        log.debug('Cells in Task: %s were reloaded', self.eid)

    def save(self, force: bool = True) -> None:
        """Save all changes in the Task

        Args:
            force: Force to update properties without digest check.

        Returns:

        """
        log.debug('Saving Task: %s...', self.eid)
        api = SignalsNotebookApi.get_default_api()

        request_body = []
        for item in self._cells:
            if item.is_changed:
                request_body.append(item.representation_for_update.dict(exclude_none=True))

        if not request_body:
            return

        api.call(
            method='PATCH',
            path=(self._get_tasks_endpoint(), self.eid, 'properties'),
            params={
                'force': json.dumps(force),
                'value': 'normalized',
            },
            json={
                'data': {'attributes': {'data': request_body}},
            },
        )
        self._reload_cells()
        log.debug('Task: %s was saved successfully', self.eid)

    def dump(
        self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None
    ) -> None:  # type: ignore[override]
        """Dump Task entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        log.debug('Dumping task: %s with name: %s...', self.eid, self.name)

        properties = [item.dict() for item in self]

        metadata = {
            'properties': [item['name'] for item in properties if item['name']],
            'filename': f'{self.name}.json',
            **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
        }
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata, default=str),
            base_alias=alias,
        )
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, f'{self.name}.json'),
            json.dumps({'data': properties}, default=str),
            base_alias=alias,
        )
        log.debug('Task: %s was dumped successfully', self.eid, self.name)

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump Task templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )

        try:
            for item in templates:
                template = cast('Task', item)
                properties = [item for item in template]
                metadata = {
                    'properties': [item.name for item in properties if item.name],
                    **{k: v for k, v in template.dict().items() if k in ('name', 'description', 'eid')},
                }
                fs_handler.write(
                    fs_handler.join_path(base_path, 'templates', entity_type, f'metadata_{template.name}.json'),
                    json.dumps(metadata),
                )
        except TypeError:
            log.exception('There is no available templates for Task entity')



import logging
from typing import Any, cast, Generator, Literal

from pydantic import Field

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import MaterialType, Response, ResponseData
from signals_notebook.materials.batch import Batch
from signals_notebook.materials.field import FieldContainer
from signals_notebook.materials.material import Material

log = logging.getLogger(__name__)


class BatchesListResponse(Response[Batch]):
    pass


class Asset(Material):
    type: Literal[MaterialType.ASSET] = Field(allow_mutation=False, default=MaterialType.ASSET)

    def __init__(self, **data: Any):
        fields = data.pop('fields', {})

        super().__init__(**data)

        self._material_fields = FieldContainer(self, self.library.asset_config.fields, **fields)

    def get_batches(self) -> Generator[Batch, None, None]:
        """Fetch batches of a specified Asset.

        Returns:
            Asset batches
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Batches for Asset: %s', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.library_name, 'assets', self.name, 'batches'),
        )

        result = BatchesListResponse(_context={'_library': self.library}, **response.json())
        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = BatchesListResponse(_context={'_library': self.library}, **response.json())
            yield from [cast(ResponseData, item).body for item in result.data]



import logging
import os

from jinja2 import Environment, FileSystemLoader, PackageLoader, select_autoescape, Template, TemplateNotFound

package_env = Environment(loader=PackageLoader('signals_notebook'), autoescape=select_autoescape())

log = logging.getLogger(__name__)


class TemplateLocationWrapper:
    """Wrapper to get template location"""

    def get_template(self, template_name: str) -> Template:
        """Get template

        Args:
            template_name: Full path to the template

        Returns:

        """
        dir_path, file_name = os.path.split(template_name)
        try:
            file_system_env = Environment(loader=FileSystemLoader(os.path.abspath(dir_path)))
            return file_system_env.get_template(file_name)
        except TemplateNotFound:
            log.info('There is no template in the system. Searching in package location...')

        try:
            return package_env.get_template(file_name)
        except TemplateNotFound:
            log.info('There is no template in the package.')
            raise


env = TemplateLocationWrapper()



from typing import Optional, Union

from pydantic import BaseModel, PrivateAttr


class SubExperimentSummaryCell(BaseModel):
    key: Optional[str]
    display: Optional[str]
    value: Optional[Union[str, int]]
    _changed: bool = PrivateAttr(default=False)

    def set_value(self, new_value: Optional[Union[str, int]]) -> None:
        self.value = new_value
        self._changed = True

    @property
    def is_changed(self) -> bool:
        return self._changed

    @property
    def representation_for_update(self):
        return {self.key: {'content': {'value': self.value}}}



import cgi
import io
import json
import logging
import time
import zipfile
from datetime import datetime
from enum import Enum
from typing import Any, cast, List, Literal, Optional, Union

import requests
from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import File, Links, MaterialType, MID, Response, ResponseData
from signals_notebook.materials.asset import Asset
from signals_notebook.materials.base_entity import BaseMaterialEntity
from signals_notebook.materials.batch import Batch
from signals_notebook.materials.field import AssetConfig, BatchConfig
from signals_notebook.utils.fs_handler import FSHandler
from signals_notebook.exceptions import SignalsNotebookError, BulkExportJobAlreadyRunningError

MAX_MATERIAL_FILE_SIZE = 52428800
EXPORT_ERROR_LIBRARY_EMPTY = 'Nothing to export.'

log = logging.getLogger(__name__)


class MaterialImportRule(str, Enum):
    NO_DUPLICATED = 'NO_DUPLICATED'
    TREAT_AS_UNIQUE = 'TREAT_AS_UNIQUE'
    USE_MATCHES = 'USE_MATCHES'


class ChangeBlameRecord(BaseModel):
    links: Links


class ChangeRecord(BaseModel):
    at: datetime = Field(allow_mutation=False)
    by: ChangeBlameRecord = Field(allow_mutation=False)

    class Config:
        validate_assignment = True


class _LibraryListData(BaseModel):
    id: str
    name: str
    digest: str
    edited: ChangeRecord
    created: ChangeRecord
    asset_config: AssetConfig = Field(alias='assets')
    batch_config: BatchConfig = Field(alias='batches')

    class Config:
        frozen = True


class LibraryListResponse(Response[_LibraryListData]):
    pass


class AssetResponse(Response[Asset]):
    pass


class BatchResponse(Response[Batch]):
    pass


class BatchAssetField(BaseModel):
    id: str
    value: Any


class BatchAssetAttribute(BaseModel):
    fields: List[BatchAssetField]


class BatchRequestData(BaseModel):
    type: str = MaterialType.BATCH
    attributes: BatchAssetAttribute


class DataRelationship(BaseModel):
    data: BatchRequestData


class AssetRelationship(BaseModel):
    batch: DataRelationship


class AssetRequestData(BaseModel):
    type: str = MaterialType.ASSET
    attributes: BatchAssetAttribute
    relationships: Optional[AssetRelationship] = None


class Library(BaseMaterialEntity):
    type: Literal[MaterialType.LIBRARY] = Field(allow_mutation=False, default=MaterialType.LIBRARY)
    _asset_config: Optional[AssetConfig] = PrivateAttr(default=None)
    _batch_config: Optional[BatchConfig] = PrivateAttr(default=None)

    class Config:
        validate_assignment = True

    def _load_configs(self) -> None:
        # the only way to get config is to fetch all libraries
        log.debug('Loading asset and batch configs to %s for %s', self.__class__.__name__, self.eid)

        result = self._get_library_list_response()
        for item in result.data:
            data = cast(_LibraryListData, cast(ResponseData, item).body)
            if data.id == self.asset_type_id:
                self._asset_config = data.asset_config
                self._batch_config = data.batch_config
                return

    @property
    def asset_config(self) -> AssetConfig:
        """Get Asset config

        Returns:
            AssetConfig
        """
        if self._asset_config:
            return self._asset_config

        self._load_configs()

        log.debug('Checking Asset Config for %s...', self.eid)
        if self._asset_config is None:
            log.warning('Asset Config for %s cannot be None', self.eid)
        assert self._asset_config is not None, f'Asset Config for {self.eid} cannot be None'
        return self._asset_config

    @asset_config.setter
    def asset_config(self, config: AssetConfig) -> None:
        """Set new AssetConfig

        Args:
            config: AssetConfig object

        Returns:

        """
        self._asset_config = config

    @property
    def batch_config(self) -> BatchConfig:
        """Get Batch config

        Returns:
            BatchConfig
        """
        if self._batch_config:
            return self._batch_config

        self._load_configs()
        log.debug('Checking Batch Config for %s...', self.eid)
        if self._batch_config is None:
            log.warning('Batch Config for %s cannot be None', self.eid)
        assert self._batch_config is not None, f'Batch Config for {self.eid} cannot be None'
        return self._batch_config

    @batch_config.setter
    def batch_config(self, config: BatchConfig) -> None:
        """Set Batch config

        Args:
            config: BatchConfig object

        Returns:

        """
        self._batch_config = config

    @classmethod
    def _get_library_list_response(cls) -> LibraryListResponse:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Library List Response for %s', cls.__name__)

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), 'libraries'),
        )

        return LibraryListResponse(**response.json())

    @classmethod
    def get_list(cls) -> List['Library']:
        """Get list of libraries

        Returns:
            list of Library objects
        """
        result = cls._get_library_list_response()
        log.debug('Get List of Libraries for %s', cls.__name__)

        libraries: List['Library'] = []
        for item in result.data:
            data = cast(_LibraryListData, cast(ResponseData, item).body)
            library = cls(
                asset_type_id=data.id,
                eid=MID(f'{MaterialType.LIBRARY}:{data.id}'),
                library_name=data.name,
                name=data.name,
                digest=data.digest.split(':')[0],
                created_at=data.created.at,
                edited_at=data.edited.at,
            )
            library.asset_config = data.asset_config
            library.batch_config = data.batch_config
            libraries.append(library)

        return libraries

    def get_asset(self, name: str) -> Asset:
        """Fetch asset from a material library by asset ID.

        Args:
            name: asset id

        Returns:
            Asset
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Asset for %s with name: %s', self.eid, name)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.name, 'assets', 'id', name),
        )

        result = AssetResponse(_context={'_library': self}, **response.json())

        return cast(ResponseData, result.data).body

    def get_asset_batches(self, name: str) -> List[Batch]:
        """Fetch batches of a specified Asset.

        Args:
            name: asset id

        Returns:
            list of Batch objects
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Asset of Batches for %s with name: %s', self.eid, name)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.name, 'assets', name, 'batches'),
        )

        result = BatchResponse(_context={'_library': self}, **response.json())

        return [cast(ResponseData, item).body for item in result.data]

    def get_batch(self, name: str) -> Batch:
        """Fetch batch from a material library by batch ID.

        Args:
            name: batch id

        Returns:
            Batch
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get a single Batch for %s with name: %s', self.eid, name)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.name, 'batches', 'id', name),
        )

        result = BatchResponse(_context={'_library': self}, **response.json())

        return cast(ResponseData, result.data).body

    def create_batch(self, asset_name: str, batch_fields: dict[str, Any]) -> Batch:
        """reate a new batch for designated asset.

        Args:
            asset_name: asset id
            batch_fields: fields of batch

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        fields = []
        log.debug('Create Batch for %s with asset name: %s', self.eid, asset_name, extra={'batch_fields': batch_fields})

        for field in self.batch_config.fields:
            if field.name in batch_fields:
                fields.append({'id': field.id, 'value': field.to_internal_value(batch_fields[field.name])})

        request_data = BatchRequestData(type='batch', attributes=BatchAssetAttribute(fields=fields))

        response = api.call(
            method='POST',
            path=(self._get_endpoint(), self.library_name, 'assets', asset_name, 'batches'),
            json={'data': request_data.dict()},
        )

        result = BatchResponse(_context={'_library': self}, **response.json())

        return cast(ResponseData, result.data).body

    def _process_asset_with_batch_fields(
        self, asset_with_batch_fields: dict[Literal[MaterialType.ASSET, MaterialType.BATCH], dict[str, Any]]
    ) -> AssetRequestData:
        request_fields = {}

        for material_instance in asset_with_batch_fields:
            request_instance_fields = []
            config: Union[AssetConfig, BatchConfig] = self.asset_config
            if material_instance == MaterialType.BATCH:
                config = self.batch_config
            for field in config.fields:
                if field.name in asset_with_batch_fields[material_instance]:
                    request_instance_fields.append(
                        {
                            'id': field.id,
                            'value': field.to_internal_value(asset_with_batch_fields[material_instance][field.name]),
                        }
                    )

            request_fields[material_instance] = request_instance_fields

        return AssetRequestData(
            type=MaterialType.ASSET,
            attributes=BatchAssetAttribute(fields=request_fields[MaterialType.ASSET]),
            relationships=AssetRelationship(
                batch=DataRelationship(
                    data=BatchRequestData(
                        type=MaterialType.BATCH,
                        attributes=BatchAssetAttribute(fields=request_fields[MaterialType.BATCH]),
                    )
                )
            ),
        )

    def create_asset_with_batches(
        self, asset_with_batch_fields: dict[Literal[MaterialType.ASSET, MaterialType.BATCH], dict[str, Any]]
    ) -> Asset:
        """Create new asset with batches

        Args:
            asset_with_batch_fields: dictionary of asset and batch fields

        Returns:
            Asset
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Create Asset with existing Batches for %s', self.eid)

        request_data = self._process_asset_with_batch_fields(asset_with_batch_fields)

        response = api.call(
            method='POST', path=(self._get_endpoint(), self.library_name, 'assets'), json={'data': request_data.dict()}
        )

        result = AssetResponse(_context={'_library': self}, **response.json())

        return cast(ResponseData, result.data).body

    def _is_file_ready(self, report_id: str) -> dict[str, Any]:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Check job status for: %s| %s', self.__class__.__name__, self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), 'bulkExport', 'reports', report_id),
        )
        response_attributes = response.json().get('data').get('attributes')
        status = response_attributes.get('status')
        error = response_attributes.get('error', None)

        result = {
            'success': response.status_code == 200 and status == 'COMPLETED',
            'error': error.get('description') if error else None,
        }

        return result

    def _download_file(self, file_id: str) -> requests.Response:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get file content for: %s| %s', self.__class__.__name__, self.eid)

        return api.call(
            method='GET',
            path=(self._get_endpoint(), 'bulkExport', 'download', file_id),
        )

    def get_content(self, timeout: int = 600, period: int = 5) -> File:
        """Get library content.
        Compounds/Reagents (SNB) will be exported to SD file, others will be exported to CSV file.

        Args:
            timeout: max available time(seconds) to get file
            period: each n seconds(default value=5) api call

        Returns:
            File
        """
        bulk_export_response = None
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get content for: %s| %s', self.__class__.__name__, self.eid)

        try:
            bulk_export_response = api.call(
                method='POST',
                path=(self._get_endpoint(), self.name, 'bulkExport'),
            )
        except SignalsNotebookError as e:
            error = e.parsed_response.errors[0]
            if error.status == '409':
                raise BulkExportJobAlreadyRunningError()
            else:
                raise
        except Exception as e:
            raise e

        file_id, report_id = bulk_export_response.json()['data']['attributes'].values()

        initial_time = time.time()

        response = None

        while time.time() - initial_time < timeout:
            result = self._is_file_ready(report_id)
            if result['error'] == EXPORT_ERROR_LIBRARY_EMPTY:
                raise FileNotFoundError('Library is empty')
            if result['success'] and not result['error']:
                response = self._download_file(file_id)
                break
            else:
                time.sleep(period)

        if not response:
            raise TimeoutError('Time is over to get file')

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'], content=response.content, content_type=response.headers.get('content-type')
        )

    def _get_import_job_completed_response(self, job_id: str) -> requests.Response:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Check job status for: %s| %s', self.__class__.__name__, self.eid)

        return api.call(method='GET', path=(self._get_endpoint(), 'bulkImport', 'jobs', job_id))

    def _import_materials(
        self,
        materials: Union[File, list[dict[Literal[MaterialType.ASSET, MaterialType.BATCH], dict[str, Any]]]],
        rule: MaterialImportRule = MaterialImportRule.TREAT_AS_UNIQUE,
        import_type: Literal['json', 'zip'] = 'json',
    ) -> requests.Response:

        api = SignalsNotebookApi.get_default_api()

        if isinstance(materials, File):
            if materials.size > MAX_MATERIAL_FILE_SIZE:
                raise ValueError('Available file size is 50Mb')

            return api.call(
                method='POST',
                path=(self._get_endpoint(), self.name, 'bulkImport'),
                params={
                    'rule': rule,
                    'importType': import_type,
                },
                headers={
                    'Content-Type': 'application/octet-stream',
                },
                data=materials.content,
            )

        request_body = [{'data': self._process_asset_with_batch_fields(material).dict()} for material in materials]

        return api.call(
            method='POST',
            path=(self._get_endpoint(), self.name, 'bulkImport'),
            params={
                'rule': rule,
                'importType': import_type,
            },
            json=request_body,
        )

    def bulk_import(  # type: ignore
        self,
        materials: Union[File, list[dict[Literal[MaterialType.ASSET, MaterialType.BATCH], dict[str, Any]]]],
        rule: MaterialImportRule = MaterialImportRule.TREAT_AS_UNIQUE,
        import_type: Literal['json', 'zip'] = 'json',
        timeout: int = 30,
        period: int = 5,
    ) -> Optional[File]:
        """Bulk import materials into a specified material library. Support import data from json or zip file.
        Zip file should contain the sdf file or csv file and attachments. T
        he column name in each records of sdf/csv file should be match the asset field name.
        In sdf file, it doesn't support character '-', '.', '<', '>', '=', '%', ' ',
        please replace them to '_' in records.
        Max materials size: 50MB.

        Rules of import:
        'TREAT_AS_UNIQUE', each item will be treated as a new asset. Selected by default.
        'USE_MATCHES', server will check the item with uniqueness check, if same will import as a batch.
        'NO_DUPLICATED', server will check the item with uniqueness check, duplicates are skipped and not imported.

        Args:
            materials: materials in zip or json format
            rule: rule of import
            import_type: import type: json or zip
            timeout: max available time(seconds) to get file
            period: each n seconds(default value=5) api call

        Returns:
            File or None
        """
        api = SignalsNotebookApi.get_default_api()

        bulk_import_response = self._import_materials(materials, rule, import_type)

        job_id = bulk_import_response.json()['data']['id']

        initial_time = time.time()

        response = False
        import_job_status = 'FAILED'

        while time.time() - initial_time < timeout:
            completed_import_response = self._get_import_job_completed_response(job_id)
            import_job_status = completed_import_response.json()['data']['attributes']['status']
            import_response_code = completed_import_response.status_code

            if import_response_code != 200 and import_job_status != 'COMPLETED':
                time.sleep(period)
            else:
                response = True
                log.debug('Library import is completed')
                break

        if not response and import_job_status == 'FAILED':
            log.debug('Time is over to import file')

            failure_report_reponse = api.call(
                method='GET',
                path=(self._get_endpoint(), 'bulkImport', 'jobs', job_id, 'failures'),
                params={
                    'filename': f'{self.name}_failure_report',
                },
            )
            content_disposition = failure_report_reponse.headers.get('content-disposition', '')
            _, params = cgi.parse_header(content_disposition)

            return File(
                name=params['filename'],
                content=failure_report_reponse.content,
                content_type=failure_report_reponse.headers.get('content-type'),
            )

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None):
        metadata = {
            **{k: v for k, v in self.dict().items() if k in ('library_name', 'asset_type_id', 'eid', 'name')},
        }
        content = self.get_content(timeout=600)
        metadata['file_name'] = content.name
        file_name = content.name
        data = content.content
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, file_name),
            data,
            base_alias=alias + [metadata['name'], file_name] if alias else None,
        )

        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            base_alias=alias + [metadata['name'], '__Metadata'] if alias else None,
        )

    @staticmethod
    def _generate_zip(my_file):
        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED, False) as zip_file:
            zip_file.writestr(f'summary.{my_file.content_type}', my_file.content)

        return zip_buffer

    def load(self, path: str, fs_handler: FSHandler):
        metadata_path = fs_handler.join_path(path, 'metadata.json')
        metadata = json.loads(fs_handler.read(metadata_path))

        content_path = fs_handler.join_path(path, metadata['file_name'])
        content_type = metadata['file_name'].split('.')[-1]
        content = fs_handler.read(content_path)

        materials = File(content=content, name='test', content_type=content_type)

        zip_buffer = self._generate_zip(materials)

        self.bulk_import(
            File(name='test', content=zip_buffer.getvalue(), content_type='zip'), import_type='zip', timeout=120
        )



import logging
from typing import Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.common_types import ObjectType
from signals_notebook.entities.plates.cell import PlateCell

log = logging.getLogger(__name__)


class PlateRow(BaseModel):
    id: Optional[UUID] = Field(allow_mutation=False, default=None)
    type: Literal[ObjectType.PLATE_ROW] = Field(allow_mutation=False, default=ObjectType.PLATE_ROW)
    cells: List[PlateCell]
    _cells_dict: Dict[Union[UUID, str], PlateCell] = PrivateAttr(default={})

    def __init__(self, **data):
        super().__init__(**data)

        for cell in self.cells:
            self._cells_dict[cell.id] = cell
            self._cells_dict[cell.name] = cell

    class Config:
        validate_assignment = True

    def __getitem__(self, index: Union[int, str, UUID]) -> PlateCell:
        if isinstance(index, int):
            return self.cells[index]

        if isinstance(index, str):
            if index in self._cells_dict:
                return self._cells_dict[index]

            try:
                if UUID(index) in self._cells_dict:
                    return self._cells_dict[UUID(index)]
            except ValueError:
                pass

        if isinstance(index, UUID):
            return self._cells_dict[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        return self.cells.__iter__()



import json
import logging
from functools import cached_property
from typing import Any, cast, ClassVar, Literal, Optional

from pydantic import BaseModel, Field

from signals_notebook.common_types import Ancestors, EntityCreationRequestPayload, EntityType, Template
from signals_notebook.entities import Notebook
from signals_notebook.entities.container import Container
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class _Attributes(BaseModel):
    name: str
    description: Optional[str] = None


class _Relationships(BaseModel):
    template: Optional[Template] = None
    ancestors: Optional[Ancestors] = None


class _RequestBody(BaseModel):
    type: EntityType
    attributes: _Attributes
    relationships: Optional[_Relationships] = None


class _RequestPayload(EntityCreationRequestPayload[_RequestBody]):
    pass


class RequestContainer(Container):
    type: Literal[EntityType.REQUEST] = Field(allow_mutation=False)
    _template_name: ClassVar = 'request.html'

    class Config:
        keep_untouched = (cached_property,)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.REQUEST

    @classmethod
    def create(
        cls,
        *,
        name: str,
        description: Optional[str] = None,
        template: Optional['RequestContainer'] = None,
        notebook: Optional[Notebook] = None,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> 'RequestContainer':
        """Create new Request in Signals Notebook

        Args:
            name: name of request
            description: description of request
            template: request template
            notebook: notebook where create request
            digest: Indicate digest
            force: Force to create without doing digest check

        Returns:
            Request
        """

        relationships = None
        if template or notebook:
            relationships = _Relationships(
                ancestors=Ancestors(data=[notebook.short_description]) if notebook else None,
                template=Template(data=template.short_description) if template else None,
            )

        request = _RequestPayload(
            data=_RequestBody(
                type=cls._get_entity_type(),
                attributes=_Attributes(
                    name=name,
                    description=description,
                ),
                relationships=relationships,
            )
        )

        log.debug('Creating Request Container for: %s', cls.__name__)
        return cast('RequestContainer', super()._create(digest=digest, force=force, request=request))

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {
            'title': self.name,
            'description': self.description,
            'edited_at': self.edited_at,
            'children': self.get_children(),
        }

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, notebook: Notebook) -> None:
        cls._load(path, fs_handler, notebook)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        metadata = json.loads(fs_handler.read(fs_handler.join_path(path, 'metadata.json')))
        request_container = cls.create(
            notebook=parent, name=metadata['name'], description=metadata['description'], force=True
        )
        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_entity_type = child_entity.split(':')[0]
            ItemMapper.get_item_class(child_entity_type)._load(
                fs_handler.join_path(path, child_entity), fs_handler, request_container
            )



import logging
from enum import Enum
from typing import Any, Dict, IO, Iterable, Mapping, Optional, Sequence, Tuple, Union

import requests

from signals_notebook.exceptions import SignalsNotebookError

log = logging.getLogger(__name__)

_Data = Union[None, str, bytes, Mapping[str, Any], Mapping[str, Any], Iterable[Tuple[str, Optional[str]]], IO[Any]]


class SignalsNotebookApi:

    _default_api_instance = None
    _api_host = ''

    API_VERSION = 'v1.0'
    """api version. Default = "v1.0" (str)
    """
    BASE_PATH = 'api/rest'
    """api base path (str)
    """
    HTTP_DEFAULT_HEADERS = {
        'Content-Type': 'application/vnd.api+json',
    }
    """ headers that are used in api (dict)
    """

    def __init__(self, session: requests.Session):
        """
        Args:
            session: A Requests session
        """
        self._session = session

    @classmethod
    def init(cls, api_host: str, api_key: str) -> 'SignalsNotebookApi':
        """Initialize SignalsNotebookApi with api host and api key

        Args:
            api_host: api host for signals notebook api
            api_key: api key for signals notebook api

        Returns:
            SignalsNotebookApi
        """
        cls._api_host = api_host

        log.info('Initialize session for api...')
        session = requests.Session()
        log.info('Session created. session.headers: %s', session.headers)

        session.headers.update({'x-api-key': api_key})

        api = cls(session)
        cls.set_default_api(api)
        log.info(
            'Default api configured. Host: %s | Base Path: %s | Version: %s ',
            api._api_host,
            api.BASE_PATH,
            api.API_VERSION,
        )
        return api

    @classmethod
    def set_default_api(cls, api: 'SignalsNotebookApi') -> None:
        """Set default api

        Args:
            api: default api

        Returns:

        """
        cls._default_api_instance = api

    @classmethod
    def get_default_api(cls) -> 'SignalsNotebookApi':
        """Get initialized API

        Returns:
            SignalsNotebookApi: Initialized API
        """
        if not cls._default_api_instance:
            log.error('You must initialize API before using')
            raise AttributeError('You must initialize API before using')
        return cls._default_api_instance

    def call(
        self,
        method: str,
        path: Union[str, Sequence[str]],
        params: Optional[Dict[str, Any]] = None,
        data: _Data = None,
        json: Optional[Union[list, Dict[str, Any]]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        """Makes an API call

        Args:
            method: The HTTP method name (e.g. 'GET').
            path: an absolute API path
            params: (optional) A mapping of request parameters where a key
                is the parameter name and its value is a string or an object
                which can be JSON-encoded.
            data: (optional) Dictionary, list of tuples, bytes, or file-like
                object to send in the body of the :class:`Request`.
            json:  (optional) A request body
            headers: (optional) A mapping of request headers where a key is the
                header name and its value is the header value.

        Returns:
            Response object
        """
        if not params:
            params = {}
        if not headers:
            headers = {}

        headers = {**self.HTTP_DEFAULT_HEADERS, **headers}

        for key, value in headers.items():
            if isinstance(value, Enum):
                headers[key] = value.value

        if json:
            response = self._session.request(
                method=method,
                url=self._prepare_path(path),
                params=params,
                json=json,
                headers=headers,
            )
        elif data:
            response = self._session.request(
                method=method,
                url=self._prepare_path(path),
                params=params,
                data=data,
                headers=headers,
            )
        else:
            response = self._session.request(
                method=method,
                url=self._prepare_path(path),
                params=params,
                headers=headers,
            )

        if not response.ok:
            log.error(
                'Error has been occurred while getting response, status code: %s',
                response.status_code,
                extra={'response': response},
            )
            raise SignalsNotebookError(response)
        log.info('Successful request - HTTP url: %s, status code: %s', response.url, response.status_code)

        return response

    @classmethod
    def _prepare_path(cls, path: Union[str, Sequence[str]]) -> str:
        if not isinstance(path, str):
            return '/'.join((cls._api_host, cls.BASE_PATH, cls.API_VERSION, *path))

        return path



import json
import logging
from enum import Enum
from functools import cached_property
from typing import Any, cast, ClassVar, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import ChemicalDrawingFormat, EntityType, File, Response, ResponseData
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.entities.stoichiometry.stoichiometry import Stoichiometry
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)

EMPTY_CDXML_FILE_CONTENT = b'<CDXML />'


class ChemicalDrawingPosition(str, Enum):
    REACTANTS = 'reactants'
    REAGENTS = 'reagents'
    PRODUCTS = 'products'


class ChemicalStructure(str, Enum):
    REACTANT = 'reactant'
    REAGENT = 'reagent'
    PRODUCT = 'product'


class ChemicalStructureFormat(str, Enum):
    INCHI = 'inchi'
    CDXML = 'cdxml'


class Structure(BaseModel):
    id: str
    type: ChemicalStructure
    inchi: Optional[str]
    cdxml: Optional[str]

    class Config:
        validate_assignment = True


class ChemicalDrawingResponse(Response[Structure]):
    pass


class StructureAttribute(BaseModel):
    dataType: ChemicalStructureFormat  # noqa
    data: str


class StructureRequestData(BaseModel):
    attributes: StructureAttribute


class ChemicalDrawing(ContentfulEntity):
    class ContentType(str, Enum):
        CDX = 'chemical/x-cdx'
        CDXML = 'chemical/x-cdxml'
        SDF = 'chemical/x-mdl-sdfile'
        MOL = 'chemical/x-mdl-molfile'
        RXN = 'chemical/x-mdl-rxnfile'
        SW = 'chemical/x-swissprot'
        SVG = 'image/svg+xml'
        CSV = 'text/csv'
        SMILES = 'chemical/x-daylight-smiles'

    class CreationContentType(str, Enum):
        CDX = 'chemical/x-cdx'
        CDXML = 'chemical/x-cdxml'
        MOL = 'chemical/x-mdl-molfile'

    type: Literal[EntityType.CHEMICAL_DRAWING] = Field(allow_mutation=False)
    _template_name: ClassVar = 'chemical_drawing.html'

    class Config:
        keep_untouched = (cached_property,)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.CHEMICAL_DRAWING

    @classmethod
    def _get_chemical_drawing_endpoint(cls) -> str:
        return 'chemicaldrawings'

    def get_structures(self, positions: ChemicalDrawingPosition) -> list[Structure]:
        """Get reactants, reagents and products of ChemicalDrawing

        Args:
            positions:  one of the ChemicalDrawing positions

        Returns:
            list of Structure objects
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Reloading structures in ChemicalDrawing: %s...', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_chemical_drawing_endpoint(), self.eid, 'reaction', positions),
        )

        result = ChemicalDrawingResponse(**response.json())

        return [cast(ResponseData, item).body for item in result.data]

    def add_structures(
        self,
        structure: Structure,
        positions: ChemicalDrawingPosition,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> Structure:
        """Add reagent, reactant or product to ChemicalDrawing

        Args:
            structure: Structure object
            positions: one of the ChemicalDrawing positions
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
                If the parameter 'force' is true, this parameter is optional.
                If the parameter 'force' is false, this parameter is required.
            force: Force to create without doing digest check

        Returns:
            Added Structure
        """
        api = SignalsNotebookApi.get_default_api()

        if structure.inchi:
            data_type = ChemicalStructureFormat.INCHI
            data = structure.inchi
        elif structure.cdxml:
            data_type = ChemicalStructureFormat.CDXML
            data = structure.cdxml
        else:
            raise ValueError('Structure doesn"t contain inchi and cdxml data')

        request_data = StructureRequestData(attributes=StructureAttribute(dataType=data_type, data=data))

        response = api.call(
            method='POST',
            path=(self._get_chemical_drawing_endpoint(), self.eid, 'reaction', positions),
            params={
                'digest': digest,
                'force': json.dumps(force),
            },
            json={'data': request_data.dict()},
        )
        result = ChemicalDrawingResponse(**response.json())
        return cast(ResponseData, result.data).body

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content_type: str = CreationContentType.CDXML,
        content: bytes = b'',
        force: bool = True,
    ) -> Entity:
        """Create ChemicalDrawing Entity

        Args:
            container: Container where create new ChemicalDrawing
            name: file name
            content_type: type of the file
            content: Entity content
            force: Force to post attachment

        Returns:
            ChemicalDrawing
        """
        if content_type:
            cls.CreationContentType(content_type)
        log.debug('Create entity: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=content,
            content_type=content_type,
            force=force,
        )

    def get_content(self, format: Optional[ChemicalDrawingFormat] = None) -> File:
        """Get Entity content

        Args:
            format: Export resource format

        Returns:

        """
        return super()._get_content(format=format)

    @cached_property
    def stoichiometry(self) -> Union[Stoichiometry, list[Stoichiometry]]:
        """Fetch stoichiometry data of ChemicalDrawing

        Returns:
            Union[Stoichiometry, list[Stoichiometry]]
        """
        return Stoichiometry.fetch_data(self.eid)

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {'name': self.name, 'stoichiometry': {}}
        file = self.get_content(format=ChemicalDrawingFormat.SVG)
        data['svg'] = 'data:{};base64,{}'.format(file.content_type, file.base64.decode('ascii'))
        if isinstance(self.stoichiometry, Stoichiometry):
            data['stoichiometry_html'] = self.stoichiometry.get_html()

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump ChemicalDrawing templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )
        try:
            for template in templates:
                template.dump(
                    fs_handler.join_path(base_path, 'templates', entity_type),
                    fs_handler,
                    ['Templates', entity_type.value],
                )
        except TypeError:
            pass

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump ChemicalDrawing

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        content = self.get_content()

        if content.content != EMPTY_CDXML_FILE_CONTENT:
            metadata: Dict[str, Any] = {
                'file_name': content.name,
                'content_type': content.content_type,
                **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
            }
            if content.content_type == self.ContentType.CDXML:
                reactants = self.get_structures(positions=ChemicalDrawingPosition.REACTANTS)
                products = self.get_structures(positions=ChemicalDrawingPosition.PRODUCTS)
                reagents = self.get_structures(positions=ChemicalDrawingPosition.REAGENTS)
                metadata = {
                    **metadata,
                    **{
                        'reactants': [{'id': i.id, 'inchi': i.inchi, 'cdxml': i.cdxml} for i in reactants],
                        'products': [{'id': i.id, 'inchi': i.inchi, 'cdxml': i.cdxml} for i in products],
                        'reagents': [{'id': i.id, 'inchi': i.inchi, 'cdxml': i.cdxml} for i in reagents],
                    },
                }

            fs_handler.write(
                fs_handler.join_path(base_path, self.eid, 'metadata.json'),
                json.dumps(metadata),
                base_alias=alias + [self.name, '__Metadata'] if alias else None
            )
            file_name = content.name
            data = content.content
            fs_handler.write(fs_handler.join_path(
                base_path, self.eid, file_name), data,
                base_alias=alias + [self.name, file_name] if alias else None
            )



import logging
from typing import cast, Union

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import MID, Response, ResponseData
from signals_notebook.materials.asset import Asset
from signals_notebook.materials.batch import Batch
from signals_notebook.materials.library import Library
from signals_notebook.materials.material import Material

log = logging.getLogger(__name__)


class MaterialResponse(Response[Union[Library, Asset, Batch]]):
    pass


class MaterialStore:
    @classmethod
    def _get_endpoint(cls) -> str:
        return 'materials'

    @classmethod
    def get(cls, eid: MID) -> Material:
        """Fetch material by entity ID.

        Args:
            eid: Unique material identifier

        Returns:
            Material
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get Material Store for %s', eid)

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), eid),
        )

        result = MaterialResponse(**response.json())

        return cast(ResponseData, result.data).body



import logging
from enum import Enum
from typing import cast, Generator, Literal, Optional

from pydantic import BaseModel
from pydantic.fields import PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import AttrID, ObjectType, Response, ResponseData

log = logging.getLogger(__name__)


class Action(str, Enum):
    UPDATE = 'update'
    DELETE = 'delete'
    CREATE = 'create'


class _Attributes(BaseModel):
    action: Action
    value: Optional[str]


class _OptionRepresentation(BaseModel):
    id: Optional[str]
    type: Literal[ObjectType.ATTRIBUTE_OPTION] = ObjectType.ATTRIBUTE_OPTION
    attributes: _Attributes


class _AttributeOption(BaseModel):
    value: str


class _AttributeOptionResponse(Response[_AttributeOption]):
    pass


class Attribute(BaseModel):
    type: str
    id: AttrID
    name: str
    _options: list[str] = PrivateAttr(default=[])

    def __init__(self, *args, **kwargs):
        self._options = kwargs.pop('options', [])
        super().__init__(*args, **kwargs)
        self.type = ObjectType.CHOICE

    def __iter__(self):
        if not self._options:
            self._reload_options()
        return self._options.__iter__()

    def __len__(self) -> int:
        if not self._options:
            self._reload_options()
        return len(self._options)

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'attributes'

    @classmethod
    def get(cls, id: AttrID) -> 'Attribute':
        """Get Attribute object by id

        Args:
            id: AttrID

        Returns:
            Attribute
        """
        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), id),
        )

        result = AttributeResponse(**response.json())
        log.debug('Get Attribute with ID: %s', id)

        return cast(ResponseData, result.data).body

    @classmethod
    def get_list(cls) -> Generator['Attribute', None, None]:
        """Get all Attributes.

        Returns:
           list of available Attributes
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get List of Attributes')

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(),),
        )
        result = AttributeResponse(**response.json())

        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = AttributeResponse(**response.json())
            yield from [cast(ResponseData, item).body for item in result.data]

        log.debug('List of Attributes was got successfully.')

    @classmethod
    def create(
        cls,
        name: str,
        description: str,
        options: Optional[list[str]] = None,
    ) -> 'Attribute':
        """Create new Attribute

        Args:
            name: name of new Attribute
            description: description of new Attribute
            options: list of available id options for created Attribute

        Returns:
            Attribute
        """
        api = SignalsNotebookApi.get_default_api()

        log.debug('Create Attribute: %s...', cls.__name__)

        response = api.call(
            method='POST',
            path=(cls._get_endpoint(),),
            json={
                'data': {
                    'type': ObjectType.ATTRIBUTE,
                    'attributes': {
                        'name': name,
                        'type': ObjectType.CHOICE,
                        'description': description,
                        'options': options or [],
                    },
                }
            },
        )
        log.debug('Attribute: %s was created.', cls.__name__)

        result = AttributeResponse(**response.json())
        return cast(ResponseData, result.data).body

    def delete(self) -> None:
        """Delete an Attribute by id.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Delete Attribute: %s...', self.id)

        api.call(
            method='DELETE',
            path=(self._get_endpoint(), self.id),
        )
        log.debug('Attribute: %s was deleted successfully', self.id)

    def add_option(self, option: str) -> None:
        """Update option of Attribute by id.

        Args:
            option: option which will be added to option list in Attribute

        Returns:

        """
        new_option = _OptionRepresentation(attributes=_Attributes(action=Action.CREATE, value=option))
        log.debug('Creating Option: %s...', self.id)
        self._patch_options(new_option)

    def delete_option(self, option: str) -> None:
        """Delete option of Attribute by id.

        Args:
            option: option which will be deleted from Attribute's option list

        Returns:

        """
        deleted_option = _OptionRepresentation(id=option, attributes=_Attributes(action=Action.DELETE))
        log.debug('Deleting Option: %s...', self.id)
        self._patch_options(deleted_option)

    def update_option(self, old_option: str, new_option: str) -> None:
        """Update option of Attribute by id.

        Args:
            old_option: old option which will be updated
            new_option: new option value

        Returns:

        """
        option = _OptionRepresentation(id=old_option, attributes=_Attributes(action=Action.UPDATE, value=new_option))
        log.debug('Patching Option: %s...', self.id)
        self._patch_options(option)

    def _patch_options(self, option: _OptionRepresentation) -> None:
        api = SignalsNotebookApi.get_default_api()
        api.call(
            method='PATCH',
            path=(self._get_endpoint(), self.id, 'options'),
            json={'data': [option.dict(exclude_none=True)]},
        )
        self._reload_options()
        log.debug('Attribute: %s was reloaded successfully', self.id)

    def _reload_options(self):
        self._options = []
        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.id, 'options'),
        )

        result = _AttributeOptionResponse(**response.json())

        for item in result.data:
            body = cast(ResponseData, item).body
            option = cast(_AttributeOption, body)

            self._options.append(option.value)

    @property
    def options(self) -> list[str]:
        """Get Attribute options

        Returns:
            list[str]
        """
        if not self._options:
            self._reload_options()
        return self._options


class AttributeResponse(Response[Attribute]):
    pass



import logging
from enum import Enum
from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType, File
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity

log = logging.getLogger(__name__)


class PowerPoint(ContentfulEntity):
    class ContentType(str, Enum):
        PPTX = 'application/vnd.openxmlformats-officedocument.presentationml.presentation'
        PPSX = 'application/vnd.openxmlformats-officedocument.presentationml.slideshow'
        POTX = 'application/vnd.openxmlformats-officedocument.presentationml.template'
        PPTM = 'application/vnd.ms-powerpoint.presentation.macroEnabled.12'
        PPSM = 'application/vnd.ms-powerpoint.slideshow.macroEnabled.12'
        POTM = 'application/vnd.ms-powerpoint.presentation.macroEnabled.12'
        PPT = 'application/vnd.ms-powerpoint'

    type: Literal[EntityType.POWER_POINT] = Field(allow_mutation=False)
    _template_name: ClassVar = 'power_point.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.POWER_POINT

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content_type: str = ContentType.PPTX,
        content: bytes = b'',
        force: bool = True,
    ) -> Entity:
        """Create PowerPoint Entity

        Args:
            container: Container where create new PowerPoint
            name: file name
            content_type: content type of PowerPoint entity
            content: PowerPoint content
            force: Force to post attachment

        Returns:
            PowerPoint
        """
        cls.ContentType(content_type)
        log.debug('Create entity: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=content,
            content_type=content_type,
            force=force,
        )

    def get_content(self) -> File:
        """Get PowerPoint content

        Returns:
            File
        """
        return super()._get_content()



import logging
from enum import Enum
from typing import Annotated, Any, List, Literal, Optional, TYPE_CHECKING, Union

from pydantic import BaseModel, Field

from signals_notebook.attributes import Attribute
from signals_notebook.common_types import AttrID, File

if TYPE_CHECKING:
    from signals_notebook.materials.material import Material

log = logging.getLogger(__name__)


class CollectionType(str, Enum):
    LIST = 'LIST'
    MULTI_SELECT = 'MULTI_SELECT'


class MaterialFieldType(str, Enum):
    ATTACHED_FILE = 'ATTACHED_FILE'
    ATTRIBUTE = 'ATTRIBUTE'
    BOOLEAN = 'BOOLEAN'
    CHEMICAL_DRAWING = 'CHEMICAL_DRAWING'
    DATETIME = 'DATETIME'
    DECIMAL = 'DECIMAL'
    EXTERNAL_LINK = 'EXTERNAL_LINK'
    LINK = 'LINK'
    MOLECULAR_FORMULA = 'MOLECULAR_FORMULA'
    MOLECULAR_MASS = 'MOLECULAR_MASS'
    TEMPERATURE = 'TEMPERATURE'
    TEXT = 'TEXT'
    CAS_NUMBER = 'CAS_NUMBER'
    DENSITY = 'DENSITY'
    SEQUENCE = 'SEQUENCE'
    SEQUENCE_FILE = 'SEQUENCE_FILE'
    INTEGER = 'INTEGER'


class BaseFieldDefinition(BaseModel):
    id: str
    name: str
    mandatory: bool
    hidden: bool
    calculated: bool = Field(default=False)
    read_only: bool = Field(alias='readOnly', default=False)
    data_type: Union[MaterialFieldType, str] = Field(alias='dataType')
    defined_by: str = Field(alias='definedBy', default='SYSTEM_DEFAULT')

    def to_internal_value(self, value: Any) -> Any:
        return value

    def to_representation(self, value: Any, material: 'Material', **kwargs) -> Any:
        return value


class TextFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.TEXT] = Field(alias='dataType', default=MaterialFieldType.TEXT)
    collection: Optional[CollectionType] = Field(default=None)
    options: Optional[List[str]] = Field(default=None)


class ChemicalDrawingFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.CHEMICAL_DRAWING] = Field(
        alias='dataType', default=MaterialFieldType.CHEMICAL_DRAWING
    )


class MolecularMassFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.MOLECULAR_MASS] = Field(
        alias='dataType', default=MaterialFieldType.MOLECULAR_MASS
    )


class DecimalFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.DECIMAL] = Field(alias='dataType', default=MaterialFieldType.DECIMAL)


class IntegerFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.INTEGER] = Field(alias='dataType', default=MaterialFieldType.INTEGER)


class BooleanFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.BOOLEAN] = Field(alias='dataType', default=MaterialFieldType.BOOLEAN)


class MolecularFormulaFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.MOLECULAR_FORMULA] = Field(
        alias='dataType', default=MaterialFieldType.MOLECULAR_FORMULA
    )


class CASNumberFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.CAS_NUMBER] = Field(alias='dataType', default=MaterialFieldType.CAS_NUMBER)


class DensityFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.DENSITY] = Field(alias='dataType', default=MaterialFieldType.DENSITY)


class AttachedFileFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.ATTACHED_FILE] = Field(
        alias='dataType', default=MaterialFieldType.ATTACHED_FILE
    )

    def to_representation(self, value: Any, material: 'Material', **kwargs) -> Any:
        """Get Material field attachment

        Args:
            value: field value
            material: material
            **kwargs:

        Returns:
            One of the Materials
        """
        return material.get_attachment(self.id)

    def to_internal_value(self, value: Any) -> Any:
        """Get field value as dictionary

        Args:
            value: field value

        Returns:
            field value as dictionary
        """
        if not isinstance(value, File):
            log.exception('File expected')
            raise TypeError('File expected')

        return {'filename': value.name, 'base64': value.base64.decode('utf-8')}


class SequenceFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.SEQUENCE] = Field(alias='dataType', default=MaterialFieldType.SEQUENCE)


class TemperatureFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.TEMPERATURE] = Field(alias='dataType', default=MaterialFieldType.TEMPERATURE)


class LinkFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.LINK] = Field(alias='dataType', default=MaterialFieldType.LINK)

    def to_representation(self, value: Any, material: 'Material', **kwargs) -> Any:
        """Get Entity by value id

        Args:
            value: field value
            material: material
            **kwargs:

        Returns:
            One of the Entities
        """
        if not value:
            return None

        from signals_notebook.entities.entity_store import EntityStore

        return EntityStore.get(value['eid'])

    def to_internal_value(self, value: Any) -> Any:
        """Get field value as dictionary

        Args:
            value: field value

        Returns:
            field value as dictionary
        """
        from signals_notebook.entities.entity import Entity

        if not isinstance(value, Entity):
            log.exception('Entity expected')
            raise TypeError('Entity expected')

        return {'eid': value.eid, 'name': value.name, 'type': value.type}


class AttributeFieldDefinition(BaseFieldDefinition):
    data_type: Literal[MaterialFieldType.ATTRIBUTE] = Field(alias='dataType', default=MaterialFieldType.ATTRIBUTE)
    multi_select: bool = Field(alias='multiSelect', default=False)
    attribute_id: AttrID = Field(alias='attribute')

    @property
    def attribute(self) -> Attribute:
        """Get Attribute object by id

        Returns:
            Attribute
        """
        return Attribute.get(self.attribute_id)


GenericFieldDefinition = Union[
    Annotated[
        Union[
            AttachedFileFieldDefinition,
            AttributeFieldDefinition,
            BooleanFieldDefinition,
            CASNumberFieldDefinition,
            ChemicalDrawingFieldDefinition,
            DecimalFieldDefinition,
            DensityFieldDefinition,
            IntegerFieldDefinition,
            LinkFieldDefinition,
            MolecularFormulaFieldDefinition,
            MolecularMassFieldDefinition,
            SequenceFieldDefinition,
            TemperatureFieldDefinition,
            TextFieldDefinition,
        ],
        Field(discriminator='data_type'),
    ],
    BaseFieldDefinition,
]


class Numbering(BaseModel):
    format: str


class AssetConfig(BaseModel):
    numbering: Numbering
    fields: List[GenericFieldDefinition]
    display_name: str = Field(alias='displayName')
    asset_name_field_id: Optional[str] = Field(alias='assetNameFieldId', default=None)

    class Config:
        frozen = True
        allow_population_by_field_name = True


class BatchConfig(BaseModel):
    numbering: Numbering
    fields: List[GenericFieldDefinition]
    display_name: str = Field(alias='displayName')

    class Config:
        frozen = True
        allow_population_by_field_name = True


class MaterialField(BaseModel):
    is_changed: bool = False
    definition: GenericFieldDefinition = Field(allow_mutation=False)
    value: Any

    class Config:
        validate_assignment = True


class FieldContainer:
    def __init__(self, material: 'Material', field_definitions: List[GenericFieldDefinition], **data):
        self._data: dict[str, MaterialField] = {}
        self._material = material

        for field_definition in field_definitions:
            self._data[field_definition.name] = MaterialField.construct(
                definition=field_definition,
                value=data.get(field_definition.name, {}).get('value'),
                is_changed=False,
            )

    def items(self):
        """Get data items

        Returns:
            A view object that displays a list of data's (key, value) tuple pair.
        """
        return self._data.items()

    def __getitem__(self, key: str) -> Any:
        field = self._data[key]

        return field.definition.to_representation(field.value, self._material)

    def __setitem__(self, key: str, value: Any):
        if key not in self._data:
            raise KeyError()

        field = self._data[key]

        field.value = field.definition.to_internal_value(value)
        field.is_changed = True



from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field


class ColumnDataType(str, Enum):
    NUMBER = 'number'
    INTEGER = 'integer'
    TEXT = 'text'
    HTML = 'html'
    BOOLEAN = 'boolean'
    UNIT = 'unit'
    LINK = 'link'


class ColumnDefinition(BaseModel):
    key: str
    title: str
    type: ColumnDataType
    measure: Optional[str]
    read_only: Optional[bool] = Field(default=None, alias='readOnly')
    hidden: Optional[bool] = Field(default=None)
    hideable: Optional[bool] = Field(default=None)
    prevent_edit_identity: Optional[bool] = Field(default=None, alias='preventEditIdentity')
    prevent_toggle_read_only: Optional[bool] = Field(default=None, alias='preventToggleReadOnly')

    class Config:
        frozen = True


class ColumnDefinitions(BaseModel):
    reactants: Optional[list[ColumnDefinition]]
    products: Optional[list[ColumnDefinition]]
    solvents: Optional[list[ColumnDefinition]]
    conditions: Optional[list[ColumnDefinition]]

    class Config:
        frozen = True



import inspect
import logging
from typing import no_type_check

from pydantic import BaseModel

log = logging.getLogger(__name__)


class PatchedModel(BaseModel):
    """
    https://github.com/samuelcolvin/pydantic/issues/1577
    Adds ability to use properties with setters
    """

    @no_type_check
    def __setattr__(self, name, value):
        try:
            super().__setattr__(name, value)
        except ValueError as e:
            log.exception('Cannot set attribute to %s', self.__class__.__name__)
            setters = inspect.getmembers(
                self.__class__, predicate=lambda x: isinstance(x, property) and x.fset is not None
            )
            for setter_name, _ in setters:
                if setter_name == name:
                    object.__setattr__(self, name, value)
                    break
            else:
                raise e



from signals_notebook.entities.entity import Entity  # noqa
from signals_notebook.entities.entity_store import EntityStore  # noqa
from signals_notebook.entities.notebook import Notebook  # noqa
from signals_notebook.entities.experiment import Experiment  # noqa
from signals_notebook.entities.text import Text  # noqa
from signals_notebook.entities.chemical_drawing import ChemicalDrawing  # noqa
from signals_notebook.common_types import ChemicalDrawingFormat  # noqa
from signals_notebook.entities.tables.table import Table  # noqa
from signals_notebook.entities.image import Image  # noqa
from signals_notebook.entities.word import Word  # noqa
from signals_notebook.entities.excel import Excel  # noqa
from signals_notebook.entities.biological_sequence import BiologicalSequence  # noqa
from signals_notebook.entities.samples.sample import Sample, SampleCell  # noqa
from signals_notebook.entities.samples.samples_container import SamplesContainer  # noqa
from signals_notebook.entities.todo_list.task import Task, TaskCell  # noqa
from signals_notebook.entities.todo_list.todo_list import TodoList  # noqa
from signals_notebook.entities.power_point import PowerPoint  # noqa
from signals_notebook.entities.spotfire import Spotfire  # noqa
from signals_notebook.entities.uploaded_resource import UploadedResource  # noqa
from signals_notebook.entities.plates.plate_container import PlateContainer  # noqa
from signals_notebook.entities.materials_table import MaterialsTable  # noqa
from signals_notebook.entities.parallel_experiment.parallel_experiment import ParallelExperiment  # noqa
from signals_notebook.entities.parallel_experiment.sub_experiment import SubExperiment  # noqa
from signals_notebook.entities.parallel_experiment.sub_experiment_summary import SubExperimentSummary  # noqa
from signals_notebook.entities.parallel_experiment.sub_experiment_layout import SubExperimentLayout  # noqa
from signals_notebook.entities.admin_defined_object import AdminDefinedObject  # noqa
from signals_notebook.entities.request import RequestContainer  # noqa
from signals_notebook.entities.todo_list.task_container import TaskContainer  # noqa



import csv
import logging
from enum import Enum
from io import StringIO
from typing import Any, cast, ClassVar, Dict, Generator, List, Literal, Optional, Union
from uuid import UUID

from pydantic import Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EID, EntityType, File, Response, ResponseData
from signals_notebook.entities import Sample
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class SamplesContainerFormat(str, Enum):
    CSV = 'csv'
    SDF = 'sdf'


class SamplesContainerResponse(Response[Sample]):
    pass


class SamplesContainer(ContentfulEntity):
    type: Literal[EntityType.SAMPLES_CONTAINER] = Field(allow_mutation=False)
    _template_name: ClassVar = 'samplesContainer.html'
    _samples: List[Sample] = PrivateAttr(default=[])
    _samples_by_id: Dict[EID, Sample] = PrivateAttr(default={})

    def __getitem__(self, index: Union[int, str, UUID, EID]) -> Sample:
        if not self._samples:
            self._reload_samples()

        if isinstance(index, int):
            return self._samples[index]

        if isinstance(index, str):
            return self._samples_by_id[EID(index)]

        if isinstance(index, EID):
            return self._samples_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._samples:
            self._reload_samples()
        return self._samples.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.SAMPLES_CONTAINER

    def get_content(self, format: Optional[SamplesContainerFormat] = None) -> File:
        """Get SamplesContainer content

        Returns:
            File
        """
        return super()._get_content(format=format)

    def _reload_samples(self) -> None:
        log.debug('Reloading samples for Samples Container: %s...', self.eid)

        self._samples = []
        self._samples_by_id = {}
        for item in self._get_samples():
            sample = cast(Sample, item)
            assert sample.eid

            self._samples.append(item)
            self._samples_by_id[sample.eid] = sample
        log.debug('Data in Samples Container: %s were reloaded', self.eid)

    def _get_samples(self) -> Generator[Sample, None, None]:
        log.debug('Getting samples for Samples Container: %s...', self.eid)
        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(super()._get_endpoint(), self.eid, 'children'),
        )

        result = SamplesContainerResponse(**response.json())

        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = SamplesContainerResponse(**response.json())
            yield from [cast(ResponseData, item).body for item in result.data]
        log.debug('Samples for SamplesContainer: %s were got successfully.', self.eid)

    def save(self, force: bool = True) -> None:
        """Save SamplesContainer content.

        Args:
            force: Force to update content without doing digest check.

        Returns:

        """
        log.debug('Saving SamplesContainer: %s...', self.eid)
        for item in self._samples:
            item.save(force=force)
        self._reload_samples()
        log.debug('SamplesContainer: %s were saved successfully.', self.eid)

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered HTML in string format
        """
        file = self.get_content()
        content = StringIO(file.content.decode('utf-8'))
        csv_data = list(csv.reader(content))
        table_head = csv_data[0]
        rows = csv_data[1:]
        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(name=self.name, table_head=table_head, rows=rows)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump SampleContainer

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        super().dump(base_path=base_path, fs_handler=fs_handler, alias=alias + [self.name] if alias else None)

        samples_path = fs_handler.join_path(base_path, self.eid)
        for item in self:
            item.dump(base_path=samples_path, fs_handler=fs_handler, alias=alias + [self.name] if alias else None)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, parent: Container) -> None:
        """Load SampleContainer

        Args:
            path: content path where create templates dump
            fs_handler: FSHandler
            parent: Container where load SampleContainer entity and Samples

        Returns:

        """
        cls._load(path, fs_handler, parent)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_entity_type = child_entity.split(':')[0]
            ItemMapper.get_item_class(child_entity_type)._load(
                fs_handler.join_path(path, child_entity), fs_handler, parent
            )



from typing import Type

from signals_notebook.entities import Entity


class ItemMapper:
    @staticmethod
    def get_item_class(item_name: str) -> Type['Entity']:
        return [
            subclass
            for subclass in Entity.get_subclasses()
            if subclass._get_entity_type() and subclass._get_entity_type().value == item_name
        ][0]



import json
import logging
from typing import Any, cast, List, Literal, Optional

from pydantic import BaseModel, Field

from signals_notebook.common_types import EntityCreationRequestPayload, EntityType
from signals_notebook.entities.container import Container
from signals_notebook.utils.fs_handler import FSHandler

log = logging.getLogger(__name__)


class _Attributes(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    organization: Optional[str] = None


class _RequestBody(BaseModel):
    type: EntityType
    attributes: _Attributes


class _RequestPayload(EntityCreationRequestPayload[_RequestBody]):
    pass


class Notebook(Container):
    type: Literal[EntityType.NOTEBOOK] = Field(allow_mutation=False)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.NOTEBOOK

    @classmethod
    def create(
        cls,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        digest: Optional[str] = None,
        force: bool = True,
        organization: Optional[str] = None,
    ) -> 'Notebook':
        """Create Notebook in Signals Notebooks

        Args:
            name: Notebook name
            description: Notebook description
            digest: Indicate digest
            force: Force to create without doing digest check
            organization: Organization
        Returns:

        """

        request = _RequestPayload(
            data=_RequestBody(
                type=cls._get_entity_type(),
                attributes=_Attributes(name=name, description=description, organization=organization),
            )
        )

        log.debug('Creating Notebook for: %s', cls.__name__)
        return cast('Notebook', super()._create(digest=digest, force=force, request=request))

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        metadata = {k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')}
        self._reload_properties()
        for prop in self._properties:
            if prop.name == 'Organization':
                metadata['Organization'] = prop.value
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            alias + [self.name, '__Metadata'] if alias else None,
        )
        for child in self.get_children(order=None):
            child.dump(base_path + '/' + self.eid, fs_handler, alias + [self.name] if alias else None)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler) -> None:
        cls._load(path, fs_handler, None)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        metadata = json.loads(fs_handler.read(fs_handler.join_path(path, 'metadata.json')))
        try:
            notebook = cls.create(
                name='restore:' + metadata['name'],
                description=metadata['description'],
                organization=metadata['Organization'],
                force=True,
            )
        except Exception as e:
            log.error(str(e))
            if 'According to template, name is auto generated, can not be specified' in str(e):
                log.error('Retrying create')
                notebook = cls.create(
                    description=metadata['description'], organization=metadata['Organization'], force=True
                )
            else:
                raise e
        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_entity_type = child_entity.split(':')[0]
            ItemMapper.get_item_class(child_entity_type)._load(
                fs_handler.join_path(path, child_entity), fs_handler, notebook
            )



import logging
from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType, File
from signals_notebook.entities.contentful_entity import ContentfulEntity

log = logging.getLogger(__name__)


class MaterialsTable(ContentfulEntity):
    type: Literal[EntityType.MATERIAL_TABLE] = Field(allow_mutation=False)
    _template_name: ClassVar = 'materials_table.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.MATERIAL_TABLE

    def get_content(self) -> File:
        return super()._get_content()



import logging
from enum import Enum
from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType, File
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity

log = logging.getLogger(__name__)


class UploadedResource(ContentfulEntity):
    class ContentType(str, Enum):
        BINARY = 'application/octet-stream'

    type: Literal[EntityType.UPLOADED_RESOURCE] = Field(allow_mutation=False)
    _template_name: ClassVar = 'uploaded_resource.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.UPLOADED_RESOURCE

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content: bytes = b'',
        content_type: str = ContentType.BINARY,
        force: bool = True,
    ) -> Entity:
        """Create UploadedResource Entity

        Args:
            container: Container where create new UploadedResource
            name: file name
            content: UploadedResource content
            content_type: UploadedResource content type
            force: Force to post attachment

        Returns:
            UploadedResource
        """
        log.debug('Create entity: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=content,
            content_type=content_type,
            force=force,
        )

    def get_content(self) -> File:
        """Get UploadedResource content

        Returns:
            File
        """
        return super()._get_content()



import json
import logging
from typing import Any, cast, Dict, List, Literal, Optional, TYPE_CHECKING, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import (
    Ancestors,
    EntityCreationRequestPayload,
    EntityShortDescription,
    EntityType,
    Response,
    ResponseData,
)
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.samples.cell import SampleCell, SampleCellContent
from signals_notebook.utils import FSHandler

if TYPE_CHECKING:
    from signals_notebook.entities import SamplesContainer

log = logging.getLogger(__name__)


class SampleCellBody(BaseModel):
    id: Optional[str]
    content: SampleCellContent = Field(default=SampleCellContent())


class _SampleAttributes(BaseModel):
    fields: Optional[List[SampleCellBody]] = []


class _SampleRelationships(BaseModel):
    template: Optional[Dict[str, EntityShortDescription]] = None
    ancestors: Optional[Ancestors] = None


class _SampleRequestBody(BaseModel):
    type: EntityType
    attributes: _SampleAttributes
    relationships: Optional[_SampleRelationships] = None


class _SampleRequestPayload(EntityCreationRequestPayload[_SampleRequestBody]):
    pass


class SampleCellsResponse(Response[SampleCell]):
    pass


class Sample(Entity):
    type: Literal[EntityType.SAMPLE] = Field(allow_mutation=False)
    _cells: List[SampleCell] = PrivateAttr(default=[])
    _cells_by_id: Dict[Union[str, UUID], SampleCell] = PrivateAttr(default={})

    def __getitem__(self, index: Union[int, str, UUID]) -> SampleCell:
        if not self._cells:
            self._reload_cells()

        if isinstance(index, int):
            return self._cells[index]

        if isinstance(index, str):
            try:
                return self._cells_by_id[index]
            except KeyError as e:
                raise e

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._cells:
            self._reload_cells()
        return self._cells.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.SAMPLE

    @classmethod
    def _get_samples_endpoint(cls) -> str:
        return 'samples'

    def _reload_cells(self) -> None:
        log.debug('Reloading cells for Sample: %s...', self.eid)
        self._cells = []
        self._cells_by_id = {}

        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(self._get_samples_endpoint(), self.eid, 'properties'),
            params={'value': 'normalized'},
        )

        result = SampleCellsResponse(**response.json())
        cells = [cast(ResponseData, item) for item in result.data]

        for item in cells:
            sample_cell = cast(SampleCell, item.body)
            sample_cell.read_only = getattr(item, 'meta', {}).get('definition', {}).get('readOnly', False)
            assert sample_cell.id

            self._cells.append(sample_cell)
            self._cells_by_id[sample_cell.id] = sample_cell
        log.debug('Cells for Sample: %s were reloaded', self.eid)

    def get_column_definitions_list(self) -> List[str]:
        if not self._cells:
            self._reload_cells()
        return [item.name for item in self]

    def save(self, force: bool = True) -> None:
        """Save Sample.

        Args:
            force: Force to update content without doing digest check.

        Returns:

        """
        log.debug('Saving Sample: %s...', self.eid)
        api = SignalsNotebookApi.get_default_api()

        request_body = []
        for item in self._cells:
            if item.is_changed:
                request_body.append(item.representation_for_update.dict(exclude_none=True))

        if not request_body:
            return

        api.call(
            method='PATCH',
            path=(self._get_samples_endpoint(), self.eid, 'properties'),
            params={
                'force': json.dumps(force),
                'value': 'normalized',
            },
            json={
                'data': {'attributes': {'data': request_body}},
            },
        )
        self._reload_cells()
        log.debug('Sample: %s were saved successfully.', self.eid)

    @classmethod
    def create(
        cls,
        *,
        cells: Optional[List[SampleCellBody]] = None,
        template: Optional['Sample'] = None,
        ancestors: Optional[List[Union[Container, 'SamplesContainer']]] = None,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> 'Sample':
        """Create Sample Entity

        Args:
            cells: Sample's data
            template: template for Sample creation
            ancestors: Container or SamplesContainer where create new Sample
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
                If the parameter 'force' is true, this parameter is optional.
                If the parameter 'force' is false, this parameter is required.
            force: Force to create without doing digest check

        Returns:
            Sample
        """
        log.debug('Create Sample: %s ', cls.__name__)

        relationships = None
        if template or ancestors:
            relationships = _SampleRelationships(
                ancestors=Ancestors(data=[item.short_description for item in ancestors]) if ancestors else None,
                template={'data': template.short_description} if template else None,
            )

        request = _SampleRequestPayload(
            data=_SampleRequestBody(
                type=cls._get_entity_type(),
                attributes=_SampleAttributes(fields=cells),
                relationships=relationships,
            )
        )

        return cast('Sample', super()._create(digest=digest, force=force, request=request))

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump Sample entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """

        metadata = {
            'filename': f'{self.name}.json',
            'columns': self.get_column_definitions_list(),
            **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
        }
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            base_alias=alias + [self.name, '__Metadata'] if alias else None,
        )
        data = [item.dict() for item in self]
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, f'{self.name}.json'),
            json.dumps({'data': data}, default=str).encode('utf-8'),
            base_alias=alias + [self.name, f'{self.name}.json'] if alias else None,
        )

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, parent: Container) -> None:
        """Load Sample entity

        Args:
            path: content path
            fs_handler: FSHandler
            parent: Container where load Sample entity

        Returns:

        """
        cls._load(path, fs_handler, parent)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.entities import EntityStore

        log.debug('Loading sample from dump...')

        entity_type = cls._get_entity_type()
        metadata_path = fs_handler.join_path(path, 'metadata.json')
        metadata = json.loads(fs_handler.read(metadata_path))

        sample_filename = metadata['filename']
        sample_data_path = fs_handler.join_path(path, sample_filename)
        sample_content = json.loads(fs_handler.read(sample_data_path))['data']

        cells = []
        for item in sample_content:
            if not item['read_only'] and item['name'] != 'Amount':
                try:
                    int(item['id'])
                    cells.append(SampleCellBody(**item))
                except ValueError:
                    pass

        column_definitions = metadata.get('columns')
        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )
        for item in templates:
            template = cast('Sample', item)
            template_column_definitions = template.get_column_definitions_list()
            if set(template_column_definitions) == set(column_definitions):
                cls.create(
                    ancestors=[parent],
                    template=template,
                    cells=cells,
                )
                break

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump Sample templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )

        for template in templates:
            template.dump(
                fs_handler.join_path(base_path, 'templates', entity_type),
                fs_handler,
                ['Templates', entity_type.value],
            )



import json
import logging
from datetime import datetime
from typing import Any, cast, ClassVar, Dict, Generator, Generic, List, Optional, Type, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr
from pydantic.generics import GenericModel

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import (
    EID,
    EntityCreationRequestPayload,
    EntityShortDescription,
    EntityType,
    Response,
    ResponseData,
)
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

ChildClass = TypeVar('ChildClass', bound='Entity')
CellValueType = TypeVar('CellValueType')

log = logging.getLogger(__name__)
MAIN_PROPERTIES = ['Name', 'Description', 'createdAt', 'editedAt']


class Property(GenericModel, Generic[CellValueType]):
    id: Optional[Union[UUID, str]]
    type: Optional[str]
    name: Optional[str]
    description: Optional[str]
    value: Optional[CellValueType]
    values: Optional[List[CellValueType]]
    _changed: bool = PrivateAttr(default=False)

    def set_value(self, new_value: CellValueType) -> None:
        """Set new value

        Args:
            new_value: new value of Property value field

        Returns:

        """
        self.value = new_value
        self._changed = True

    @property
    def is_changed(self) -> bool:
        """Checking if content of Cell has been modified

        Returns:
            bool
        """
        return self._changed

    @property
    def representation_for_update(self) -> dict[str, dict]:
        """Get representation of body for update

        Returns:
            dict[str, dict]
        """
        return {'attributes': self.dict(include={'name', 'value'})}


class PropertiesResponse(Response[Property]):
    pass


class Entity(BaseModel):
    type: str = Field(allow_mutation=False)
    eid: EID = Field(allow_mutation=False)
    digest: Optional[str] = Field(allow_mutation=False, default=None)
    name: str = Field(title='Name')
    description: Optional[str] = Field(title='Description', default=None)
    created_at: datetime = Field(alias='createdAt', allow_mutation=False)
    edited_at: datetime = Field(alias='editedAt', allow_mutation=False)
    _template_name: ClassVar = 'entity.html'
    _properties: List[Property] = PrivateAttr(default=[])
    _properties_by_id: Dict[Union[str, UUID], Property] = PrivateAttr(default={})

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} eid={self.eid}>'

    def __getitem__(self, index: Union[int, str, UUID, EID]) -> Any:
        if not self._properties:
            self._reload_properties()

        if isinstance(index, int):
            return self._properties[index]

        if isinstance(index, str):
            try:
                return self._properties_by_id[UUID(index)]
            except ValueError:
                return self._properties_by_id[index]

        if isinstance(index, UUID):
            return self._properties_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._properties:
            self._reload_properties()
        return self._properties.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        raise NotImplementedError

    @classmethod
    def get_subclasses(cls) -> Generator[Type['Entity'], None, None]:
        """Get all Entity subclasses

        Returns:
            One of the Entity subclasses
        """
        log.debug('Get subclasses for: %s', cls.__name__)
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @classmethod
    def set_template_name(cls, template_name: str) -> None:
        """Set name of the template

        Args:
            template_name: template name

        Returns:

        """
        log.debug('Setting new template for: %s...', cls.__name__)
        cls._template_name = template_name
        log.debug('New template (%s) for %s was set', template_name, cls.__name__)

    @classmethod
    def get_template_name(cls) -> str:
        """Get Entity template name

        Returns:
            Template name
        """
        return cls._template_name

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'entities'

    @classmethod
    def _get_list_params(cls) -> Dict[str, Any]:
        return {
            'include_types': [cls._get_entity_type()],
        }

    def _reload_properties(self) -> None:
        log.debug('Reloading properties in Entity: %s...', self.eid)
        self._properties = []
        self._properties_by_id = {}

        api = SignalsNotebookApi.get_default_api()

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'properties'),
        )
        result = PropertiesResponse(**response.json())
        properties = [cast(ResponseData, item).body for item in result.data]

        for item in properties:
            entity_property = cast(Property, item)
            if entity_property.name in MAIN_PROPERTIES:
                continue
            assert entity_property.id

            self._properties.append(entity_property)
            self._properties_by_id[entity_property.id] = entity_property
        log.debug('Properties in Entity: %s were reloaded', self.eid)

    @classmethod
    def get_list(cls) -> Generator['Entity', None, None]:
        """Get all entities

        Returns:
            list of entities
        """
        from signals_notebook.entities.entity_store import EntityStore

        return EntityStore.get_list(**cls._get_list_params())

    def delete(self) -> None:
        """Delete entity

        Returns:

        """
        from signals_notebook.entities.entity_store import EntityStore

        EntityStore.delete(self.eid)
        log.debug('Entity: %s was deleted from EntityStore', self.eid)

    @classmethod
    def _create(
        cls, *, digest: Optional[str] = None, force: bool = True, request: EntityCreationRequestPayload,
    ) -> 'Entity':
        api = SignalsNotebookApi.get_default_api()
        log.debug('Create Entity: %s...', cls.__name__)

        response = api.call(
            method='POST',
            path=(cls._get_endpoint(),),
            params={
                'digest': digest,
                'force': json.dumps(force),
            },
            json=request.dict(exclude_none=True, by_alias=True),
        )
        log.debug('Entity: %s was created.', cls.__name__)

        result = Response[cls](**response.json())  # type: ignore

        return cast(ResponseData, result.data).body

    def refresh(self) -> None:
        """Refresh entity with new changes values

        Returns:

        """
        from signals_notebook.entities import EntityStore

        EntityStore.refresh(self)

    def _patch_properties(self, request_body, force: bool) -> None:
        api = SignalsNotebookApi.get_default_api()
        api.call(
            method='PATCH',
            path=(self._get_endpoint(), self.eid, 'properties'),
            params={
                'digest': None if force else self.digest,
                'force': json.dumps(force),
            },
            json={
                'data': request_body,
            },
        )

    def save(self, force: bool = True) -> None:
        """Update attributes and properties of a specified entity.

        Args:
            force: Force to update attributes and properties without doing digest check.

        Returns:

        """
        log.debug('Save Entity: %s...', self.eid)

        request_body = []
        for field in self.__fields__.values():
            if field.field_info.allow_mutation:
                request_body.append(
                    {'attributes': {'name': field.field_info.title, 'value': getattr(self, field.name)}}
                )

        log.debug('Updating properties in Entity: %s...', self.eid)
        if self._properties:
            for item in self._properties:
                if item.is_changed:
                    request_body.append(item.representation_for_update)
        self._patch_properties(request_body=request_body, force=force)
        self._reload_properties()

        log.debug('Properties in Entity: %s were updated successfully', self.eid)
        log.debug('Entity: %s was saved.', self.eid)

    @property
    def short_description(self) -> EntityShortDescription:
        """Return EntityShortDescription of Entity

        Returns:
            EntityShortDescription
        """
        return EntityShortDescription(type=self.type, id=self.eid)

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered HTML in string format
        """
        data = {'name': self.name, 'edited_at': self.edited_at, 'type': self.type, 'description': self.description}
        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        log.error('Dumping is not implemented')

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        log.error('Loading is not implemented')

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump Entity templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )
        try:
            for template in templates:
                fs_handler.write(
                    fs_handler.join_path(base_path, 'templates', entity_type, f'metadata_{template.name}.json'),
                    json.dumps({k: v for k, v in template.dict().items() if k in ('name', 'description', 'eid')}),
                    base_alias=['Templates', entity_type, template.name],
                )

        except TypeError:
            pass



import logging
from decimal import Decimal
from enum import Enum
from typing import Any, cast, Generic, Optional, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr
from pydantic.generics import GenericModel

from signals_notebook.entities.stoichiometry.cell import ColumnDefinition
from signals_notebook.jinja_env import env

log = logging.getLogger(__name__)


class DataGridKind(str, Enum):
    REACTANTS = 'reactants'
    PRODUCTS = 'products'
    SOLVENTS = 'solvents'
    CONDITIONS = 'conditions'


class Cell(BaseModel):
    value: Union[bool, Decimal, str]
    display: Optional[str] = Field(default='')
    units: Optional[str]

    class Config:
        frozen = True


class Row(BaseModel):
    row_id: Union[UUID, str]

    class Config:
        frozen = True


RowClass = TypeVar('RowClass', bound=Row)


class Rows(GenericModel, Generic[RowClass]):
    __root__: list[RowClass]
    _rows_by_id: dict[Union[str, UUID], RowClass] = PrivateAttr(default={})
    _column_definitions: list[ColumnDefinition] = PrivateAttr(default=[])
    _template_name: str = 'data_grid.html'

    def __init__(self, **data: Any):
        """Stoichiometry Row

        Args:
            data: data to initialize Rows class
        """
        super(Rows, self).__init__(**data)

        for row in self.__root__:
            self._rows_by_id[row.row_id] = row

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, index: Union[int, str, UUID]) -> RowClass:
        if isinstance(index, int):
            return self.__root__[index]

        if isinstance(index, str):
            try:
                return self._rows_by_id[index]
            except KeyError:
                log.exception('Row will be returned by UUID index')
                return self._rows_by_id[UUID(index)]

        if isinstance(index, UUID):
            return self._rows_by_id[index]

        raise IndexError('Invalid index')

    @property
    def column_definitions(self) -> list[ColumnDefinition]:
        """Get column_definitions

        Returns:
            list of ColumnDefinition objects
        """
        return self._column_definitions

    def set_column_definitions(self, value: list[ColumnDefinition]) -> None:
        """Set column definitions to value

        Args:
            value: list of ColumnDefinition objects

        Returns:

        """
        self._column_definitions = value

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        rows = []
        for row in self.__root__:
            reformatted_row = {}

            for column_definition in self.column_definitions:
                if hasattr(row, column_definition.key) and not column_definition.hidden:
                    cell = getattr(row, column_definition.key, None)
                    cell = cast(Cell, cell)
                    reformatted_row[column_definition.title] = '' if cell is None else (cell.display or cell.value)

            rows.append(reformatted_row)

        template = env.get_template(self._template_name)
        log.info('Html template for %s has been rendered.', self.__class__.__name__)

        return template.render(rows=rows)


class Reactant(Row):
    rxnid: Cell
    name: Optional[Cell]
    cas: Optional[Cell]
    formula: Cell
    mf: Cell
    mw: Cell
    em: Cell
    limit: Optional[Cell]
    eq: Optional[Cell]
    sm: Optional[Cell]
    moles: Optional[Cell]
    loading: Optional[Cell]
    molarity: Optional[Cell]
    volume: Optional[Cell]
    density: Optional[Cell]
    wt: Optional[Cell]
    supplier: Optional[Cell]
    lot: Optional[Cell]
    iupac_name: Cell = Field(alias='IUPACName')
    barcode: Optional[Cell]
    material_id: Optional[Cell] = Field(alias='materialId')
    container_id: Optional[Cell] = Field(alias='containerId')
    inchi: Optional[str]
    hash: Optional[str]

    class Config:
        frozen = True


class Reactants(Rows[Reactant]):
    pass


class Product(Row):
    rxnid: Cell
    product_id: Cell = Field(alias='productId')
    name: Optional[Cell]
    formula: Cell
    mf: Cell
    mw: Cell
    em: Cell
    theo_mass: Optional[Cell] = Field(alias='theoMass')
    actual_mass: Optional[Cell] = Field(alias='actualMass')
    purity: Optional[Cell]
    density: Optional[Cell]
    yield_: Optional[Cell] = Field(alias='yield')
    theo_mol: Optional[Cell] = Field(alias='theoMol')
    actual_mol: Optional[Cell] = Field(alias='actualMol')
    conversion: Optional[Cell]
    iupac_name: Cell = Field(alias='IUPACName')
    barcode: Optional[Cell]
    material_id: Optional[Cell] = Field(alias='materialId')
    container_id: Optional[Cell] = Field(alias='containerId')
    inchi: Optional[str]
    hash: Optional[str]

    class Config:
        frozen = True


class Products(Rows[Product]):
    pass


class Solvent(Row):
    solvent: Cell
    ratio: Cell
    volume: Optional[Cell]

    class Config:
        frozen = True


class Solvents(Rows[Solvent]):
    pass


class Condition(Row):
    pressure: Optional[Cell]
    temperature: Optional[Cell]
    duration: Optional[Cell]

    class Config:
        frozen = True


class Conditions(Rows[Condition]):
    pass


class DataGrids(BaseModel):
    reactants: Reactants
    products: Products
    solvents: Solvents
    conditions: Conditions



import json
import logging
from functools import cached_property
from typing import Any, cast, Generator, Literal, Optional

from pydantic import BaseModel, Field

from signals_notebook.common_types import (
    Ancestors,
    EntityCreationRequestPayload,
    EntityType,
    Template,
)
from signals_notebook.entities import Entity
from signals_notebook.entities.chemical_drawing import (
    ChemicalDrawing,
    ChemicalDrawingPosition,
    ChemicalStructure,
    Structure,
)
from signals_notebook.entities.container import Container
from signals_notebook.entities.parallel_experiment.parallel_experiment import ParallelExperiment
from signals_notebook.jinja_env import env
from signals_notebook.utils.fs_handler import FSHandler

log = logging.getLogger(__name__)


class _SubExperimentAttributes(BaseModel):
    description: Optional[str] = None


class _SubExperimentRelationships(BaseModel):
    template: Optional[Template] = None
    ancestors: Optional[Ancestors] = None


class _SubExperimentRequestBody(BaseModel):
    type: EntityType
    attributes: Optional[_SubExperimentAttributes]
    relationships: Optional[_SubExperimentRelationships] = None


class _SubExperimentRequestPayload(EntityCreationRequestPayload[_SubExperimentRequestBody]):
    pass


class SubExperiment(Container):
    type: Literal[EntityType.SUB_EXPERIMENT] = Field(allow_mutation=False)
    _template_name = 'sub_experiment.html'

    class Config:
        keep_untouched = (cached_property,)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.SUB_EXPERIMENT

    def get_children(self, order='') -> Generator[Entity, None, None]:
        """Get children of SubExperiment.

        Returns:
            list of Entities
        """
        return super().get_children(order=order)

    @classmethod
    def create(
        cls,
        *,
        parallel_experiment: ParallelExperiment,
        description: Optional[str] = None,
        template: Optional['SubExperiment'] = None,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> 'SubExperiment':

        relationships = None
        if template or parallel_experiment:
            relationships = _SubExperimentRelationships(
                ancestors=Ancestors(data=[parallel_experiment.short_description]) if parallel_experiment else None,
                template=Template(data=template.short_description) if template else None,
            )

        request = _SubExperimentRequestPayload(
            data=_SubExperimentRequestBody(
                type=cls._get_entity_type(),
                attributes=_SubExperimentAttributes(description=description),
                relationships=relationships,
            )
        )

        return cast('SubExperiment', super()._create(digest=digest, force=force, request=request))

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {
            'title': self.name,
            'description': self.description,
            'edited_at': self.edited_at,
            'children': self.get_children(),
        }

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    @staticmethod
    def _add_structures_from_metadata(
        metadata: list[dict[str, str]],
        chemical_structure: ChemicalStructure,
        position: ChemicalDrawingPosition,
        chemical_drawing: ChemicalDrawing,
    ) -> None:
        for metadata_structure in metadata:
            structure = Structure(
                id=metadata_structure['id'],
                type=chemical_structure,
                inchi=metadata_structure['inchi'],
                cdxml=metadata_structure['cdxml'],
            )
            chemical_drawing.add_structures(structure=structure, positions=position)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, parallel_experiment: ParallelExperiment) -> None:
        cls._load(path, fs_handler, parallel_experiment)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        metadata = json.loads(fs_handler.read(fs_handler.join_path(path, 'metadata.json')))
        sub_experiment = cls.create(parallel_experiment=parent, description=metadata['description'], force=True)
        existing_chemical_drawing = [
            cast(ChemicalDrawing, i) for i in sub_experiment.get_children() if i.type == EntityType.CHEMICAL_DRAWING
        ][0]

        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_path = fs_handler.join_path(path, child_entity)
            metadata = json.loads(fs_handler.read(fs_handler.join_path(child_path, 'metadata.json')))
            child_entity_type = child_entity.split(':')[0]
            entity_type = ItemMapper.get_item_class(child_entity_type)

            if child_entity_type == EntityType.CHEMICAL_DRAWING:

                if not metadata.get('reactants') or not metadata.get('products'):
                    continue
                cls._add_structures_from_metadata(
                    metadata['reactants'],
                    ChemicalStructure.REACTANT,
                    ChemicalDrawingPosition.REACTANTS,
                    existing_chemical_drawing,
                )
                cls._add_structures_from_metadata(
                    metadata['products'],
                    ChemicalStructure.PRODUCT,
                    ChemicalDrawingPosition.PRODUCTS,
                    existing_chemical_drawing,
                )
                cls._add_structures_from_metadata(
                    metadata['reagents'],
                    ChemicalStructure.REAGENT,
                    ChemicalDrawingPosition.REAGENTS,
                    existing_chemical_drawing,
                )

            else:
                entity_type._load(fs_handler.join_path(path, child_entity), fs_handler, sub_experiment)



import logging
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.common_types import ObjectType
from signals_notebook.entities.parallel_experiment.cell import SubExperimentSummaryCell

log = logging.getLogger(__name__)


class Content(BaseModel):
    columns: Optional[dict]


class RowRequestBody(BaseModel):
    id: Optional[Union[UUID, str]]
    type: Literal[ObjectType.SUB_EXPERIMENT] = Field(allow_mutation=False, default=ObjectType.SUB_EXPERIMENT)
    attributes: Content

    class Config:
        validate_assignment = True


class Row(BaseModel):
    id: Optional[Union[UUID, str]]
    cells: List[SubExperimentSummaryCell] = Field(alias='columns')
    _cells_dict: Dict[Union[UUID, str], SubExperimentSummaryCell] = PrivateAttr(default={})

    class Config:
        validate_assignment = True

    def __init__(self, **data):
        super().__init__(**data)

        for cell in self.cells:
            self._cells_dict[cell.key] = cell

    def get(self, value: str, default: Any = None) -> Union[SubExperimentSummaryCell, Any]:
        """Get one of the cells

        Args:
            value: key to get one of the cells
            default: default value if key doesn't exist

        Returns:
            Union[Cell, Any]
        """
        try:
            return self[value]
        except KeyError:
            log.debug('KeyError were caught. Default value returned')
            return default

    @property
    def is_changed(self) -> bool:
        """Get is_changed field

        Returns:
            bool: True/False
        """
        return any([cell.is_changed for cell in self.cells])

    def __getitem__(self, index: str) -> SubExperimentSummaryCell:
        if isinstance(index, str):
            if index in self._cells_dict:
                return self._cells_dict[index]

        log.exception('IndexError were caught. Invalid index')
        raise IndexError('Invalid index')

    def __iter__(self):
        return self.cells.__iter__()

    @property
    def representation_for_update(self) -> List[Dict[str, str]]:
        """Present Row object as a request body to update values

        Returns:
            list of changed cells
        """
        changed_cells = []
        for cell in self.cells:
            if cell.is_changed:
                body = RowRequestBody(id=str(self.id), attributes=Content(columns=cell.representation_for_update))
                changed_cells.append(body.dict())

        return changed_cells



import logging
from typing import cast, ClassVar, Dict, List, Literal, Union
from uuid import UUID

from pydantic import Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EntityType, File, Response, ResponseData
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.entities.plates.plate_row import PlateRow
from signals_notebook.jinja_env import env

log = logging.getLogger(__name__)


class PlateContainerDataResponse(Response[PlateRow]):
    pass


class PlateContainer(ContentfulEntity):
    type: Literal[EntityType.PLATE_CONTAINER] = Field(allow_mutation=False)
    _rows: List[PlateRow] = PrivateAttr(default=[])
    _rows_by_id: Dict[UUID, PlateRow] = PrivateAttr(default={})
    _template_name: ClassVar = 'plate_container.html'

    def __getitem__(self, index: Union[int, str, UUID]) -> PlateRow:
        if not self._rows:
            self._reload_data()

        if isinstance(index, int):
            return self._rows[index]

        if isinstance(index, str):
            return self._rows_by_id[UUID(index)]

        if isinstance(index, UUID):
            return self._rows_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._rows:
            self._reload_data()

        return self._rows.__iter__()

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.PLATE_CONTAINER

    def _reload_data(self):
        api = SignalsNotebookApi.get_default_api()
        log.debug('Reloading rows for Plate Container: %s...', self.eid)

        response = api.call(
            method='GET',
            path=('plates', self.eid, 'summary'),
        )
        result = PlateContainerDataResponse(**response.json())
        self._rows = []
        self._rows_by_id = {}
        for item in result.data:
            row = cast(PlateRow, cast(ResponseData, item).body)
            assert row.id

            self._rows.append(row)
            self._rows_by_id[row.id] = row
        log.debug('Data in Plate Container: %s were reloaded', self.eid)

    def get_content(self) -> File:
        """Get PlateContainer content

        Returns:
            File
        """
        return super()._get_content()

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered HTML in string format
        """
        if not self._rows:
            self._reload_data()
        table_head = None
        if self._rows:
            table_head = self._rows[0]

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(name=self.name, table_head=table_head, rows=self._rows)
